"""
Show how to use DDP, Horovod and DeepSpeed strategies interchangeably
with a simple neural network trained on MNIST dataset.
"""
from typing import Tuple
import argparse
import sys
import time
from timeit import default_timer as timer

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision import datasets, transforms
from torch.utils.data import Dataset

import horovod.torch as hvd

import deepspeed

from itwinai.torch.distributed import (
    distributed_resources_available,
    TorchDistributedStrategy,
    TorchDDPStrategy,
    HorovodStrategy,
    DeepSpeedStrategy,
    NonDistributedStrategy
)
from itwinai.parser import ArgumentParser as ItAIArgumentParser
from itwinai.torch.reproducibility import (
    seed_worker, set_seed
)


def parse_params() -> argparse.Namespace:
    """
    Parse CLI args, which can also be loaded from a configuration file
    using the --config flag:

    >>> train.py --strategy ddp --config config.yaml
    """
    parser = ItAIArgumentParser(description='PyTorch MNIST Example')

    # Distributed ML strategy
    parser.add_argument(
        "--strategy", "-s", type=str,
        choices=['ddp', 'horovod', 'deepspeed'],
        default='ddp'
    )

    # Data and logging
    parser.add_argument('--data-dir', default='./',
                        help=('location of the training dataset in the local '
                              'filesystem'))
    parser.add_argument('--log-int', type=int, default=10,
                        help='log interval per training')
    parser.add_argument('--verbose',
                        action=argparse.BooleanOptionalAction,
                        help='Print parsed arguments')
    parser.add_argument('--restart-int', type=int, default=10,
                        help='restart interval per epoch (default: 10)')
    parser.add_argument('--download-only',
                        action=argparse.BooleanOptionalAction,
                        help='Download dataset and exit')
    parser.add_argument('--dataset-replication', type=int, default=100,
                        help='concatenate MNIST to this factor (default: 100)')
    parser.add_argument('--shuff', action='store_true', default=False,
                        help='shuffle dataset (default: False)')
    parser.add_argument('--nworker', type=int, default=0,
                        help=('number of workers in DataLoader (default: 0 -'
                              ' only main)'))
    parser.add_argument('--prefetch', type=int, default=2,
                        help='prefetch data in DataLoader (default: 2)')

    # Model
    parser.add_argument('--batch-size', type=int, default=64,
                        help='input batch size for training (default: 64)')
    parser.add_argument('--epochs', type=int, default=10,
                        help='number of epochs to train (default: 10)')
    parser.add_argument('--lr', type=float, default=0.01,
                        help='learning rate (default: 0.01)')
    parser.add_argument('--momentum', type=float, default=0.5,
                        help='momentum in SGD optimizer (default: 0.5)')

    # Reproducibility
    parser.add_argument('--rnd-seed', type=int, default=0,
                        help='seed integer for reproducibility (default: 0)')

    # Distributed ML
    parser.add_argument('--backend', type=str, default='nccl',
                        help='backend for parrallelisation (default: nccl)')
    parser.add_argument('--local_rank', type=int, default=-1,
                        help='local rank passed from distributed launcher')

    # Horovod: ignored when not using Horovod
    parser.add_argument('--fp16-allreduce', action='store_true', default=False,
                        help='use fp16 compression during allreduce')
    parser.add_argument('--use-adasum', action='store_true', default=False,
                        help='use adasum algorithm to do reduction')
    parser.add_argument('--gradient-predivide-factor', type=float, default=1.0,
                        help=('apply gradient pre-divide factor in optimizer '
                              '(default: 1.0)'))

    # DeepSpeed
    parser = deepspeed.add_config_arguments(parser)
    args = parser.parse_args()

    if args.verbose:
        args_list = [f"{key}: {val}" for key, val in args.items()]
        print("PARSED ARGS:\n", '\n'.join(args_list))

    return args


class Net(nn.Module):
    """
    Simple neural network classifier for MNIST images.
    """

    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=-1)


def train(
    model, train_loader, optimizer, epoch,
    strategy: TorchDistributedStrategy, args
):
    """
    Training function, representing an epoch.
    """
    model.train()
    t_list = []
    loss_acc = 0
    if strategy.is_main_worker:
        print("\n")
    for batch_idx, (data, target) in enumerate(train_loader):
        t = timer()
        data = data.to(strategy.device())
        target = target.to(strategy.device())
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
        if (strategy.is_main_worker and args.log_int > 0
                and batch_idx % args.log_int == 0):
            dl_size = len(train_loader.dataset)//strategy.global_world_size()
            print(
                f'Train epoch: {epoch} '
                f'[{batch_idx * len(data)}/{dl_size} '
                f'({100.0 * batch_idx / len(train_loader):.0f}%)]\t\t'
                f'Loss: {loss.item():.6f}')
        t_list.append(timer() - t)
        loss_acc += loss.item()
    if strategy.is_main_worker:
        print('TIMER: train time', sum(t_list) / len(t_list), 's')
    return loss_acc


def test(model, test_loader, strategy: TorchDistributedStrategy):
    """
    Model validation.
    """
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data = data.to(strategy.device())
            target = target.to(strategy.device())
            output = model(data)
            # Sum up batch loss
            test_loss += F.nll_loss(output, target, reduction="sum").item()
            # Get the index of the max log-probability
            pred = output.argmax(dim=1, keepdim=True)
            correct += pred.eq(target.view_as(pred)).sum().item()
    test_loss /= len(test_loader.dataset)
    if strategy.is_main_worker:
        dl_size = len(test_loader.dataset)//strategy.global_world_size()
        print(
            f'Test set: average loss: {test_loss:.4f}\t'
            f'accurate samples: {correct}/{dl_size}')
    acc_test = (
        100.0 * correct * strategy.global_world_size()
        / len(test_loader.dataset)
    )
    return acc_test


def download_mnist():
    """
    Use built-in torch datasets functions to pull MNIST dataset.
    """

    _ = datasets.MNIST(
        args.data_dir, train=True, download=True,
        transform=transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,))
        ]))
    _ = datasets.MNIST(
        args.data_dir, train=False, download=True,
        transform=transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,))
        ]))


def mnist_dataset(dataset_replication: int = 1) -> Tuple[Dataset, Dataset]:
    """Load MNIST train and test datasets, replicating them.

    Args:
        dataset_replication (int): dataset replication factor. Default 1.

    Returns:
        Tuple[Dataset, Dataset]: train dataset and test dataset.
    """
    replicated_data = [
        datasets.MNIST(args.data_dir, train=True, download=False,
                       transform=transforms.Compose([
                           transforms.ToTensor(),
                           transforms.Normalize((0.1307,), (0.3081,))
                       ]))
        for _ in range(dataset_replication)
    ]
    train_dataset = torch.utils.data.ConcatDataset(replicated_data)

    replicated_data = [
        datasets.MNIST(args.data_dir, train=False, download=False,
                       transform=transforms.Compose([
                           transforms.ToTensor(),
                           transforms.Normalize((0.1307,), (0.3081,))
                       ]))
        for _ in range(dataset_replication)
    ]
    test_dataset = torch.utils.data.ConcatDataset(replicated_data)
    return train_dataset, test_dataset


if __name__ == "__main__":

    args = parse_params()

    if args.download_only:
        # Download datasets from a location with internet access and exit.
        # This is convenient when submitting training jobs to
        # a batch system where worker nodes have no internet
        # access, like in some HPCs.
        download_mnist()
        sys.exit()

    # Instantiate Strategy
    if not distributed_resources_available():
        print("WARNING: falling back to non-distributed strategy.")
        strategy = NonDistributedStrategy()
        distribute_kwargs = {}
    elif args.strategy == 'ddp':
        strategy = TorchDDPStrategy(backend=args.backend)
        distribute_kwargs = {}
    elif args.strategy == 'horovod':
        strategy = HorovodStrategy()
        distribute_kwargs = dict(
            compression=(
                hvd.Compression.fp16 if args.fp16_allreduce
                else hvd.Compression.none
            ),
            op=hvd.Adasum if args.use_adasum else hvd.Average,
            gradient_predivide_factor=args.gradient_predivide_factor
        )
    elif args.strategy == 'deepspeed':
        strategy = DeepSpeedStrategy(backend=args.backend)
        distribute_kwargs = dict(
            config_params=dict(train_micro_batch_size_per_gpu=args.batch_size)
        )
    else:
        raise NotImplementedError(
            f"Strategy {args.strategy} is not recognized/implemented.")

    # Initialize strategy
    strategy.init()

    # Start the timer for profiling
    st = timer()

    # Set random seed for reproducibility
    torch_prng = set_seed(args.rnd_seed)

    if strategy.is_main_worker:
        print('TIMER: initialise:', timer()-st, 's')
        print('DEBUG: local ranks:', strategy.local_world_size(),
              '/ global ranks:', strategy.global_world_size())
        print('DEBUG: sys.version:', sys.version)
        print('DEBUG: args.data_dir:', args.data_dir)
        print('DEBUG: args.log_int:', args.log_int)
        print('DEBUG: args.nworker:', args.nworker)
        print('DEBUG: args.prefetch:', args.prefetch)
        print('DEBUG: args.batch_size:', args.batch_size)
        print('DEBUG: args.epochs:', args.epochs)
        print('DEBUG: args.lr:', args.lr)
        print('DEBUG: args.momentum:', args.momentum)
        print('DEBUG: args.shuff:', args.shuff)
        print('DEBUG: args.rnd_seed:', args.rnd_seed)
        print('DEBUG: args.backend:', args.backend)

    # Dataset
    train_dataset, test_dataset = mnist_dataset(args.dataset_replication)
    # Distributed dataloaders
    train_loader = strategy.create_dataloader(
        train_dataset, batch_size=args.batch_size,
        num_workers=args.nworker, pin_memory=True,
        persistent_workers=(args.nworker > 1),
        prefetch_factor=args.prefetch, generator=torch_prng,
        worker_init_fn=seed_worker
    )
    test_loader = strategy.create_dataloader(
        test_dataset, batch_size=args.batch_size,
        num_workers=args.nworker, pin_memory=True,
        persistent_workers=(args.nworker > 1),
        prefetch_factor=args.prefetch, generator=torch_prng,
        worker_init_fn=seed_worker
    )

    if strategy.is_main_worker:
        print('TIMER: read and concat data:', timer()-st, 's')

    # Create CNN model
    model = Net().to(strategy.device())

    # Optimizer
    optimizer = torch.optim.SGD(
        model.parameters(), lr=args.lr, momentum=args.momentum)

    # Distributed model, optimizer, and scheduler
    model, optimizer, _ = strategy.distributed(
        model, optimizer, lr_scheduler=None, **distribute_kwargs
    )

    # Start training and test loop
    if strategy.is_main_worker:
        print('TIMER: broadcast:', timer()-st, 's')
        print('\nDEBUG: start training')
        print('--------------------------------------------------------')

    et = timer()
    start_epoch = 1
    for epoch in range(start_epoch, args.epochs + 1):
        lt = timer()
        if strategy.is_distributed:
            # Inform the sampler that a new epoch started: shuffle
            # may be needed
            train_loader.sampler.set_epoch(epoch)
            test_loader.sampler.set_epoch(epoch)

        # Training
        loss_acc = train(
            model=model,
            train_loader=train_loader,
            optimizer=optimizer,
            epoch=epoch,
            strategy=strategy,
            args=args
        )

        # Testing
        acc_test = test(
            model=model,
            test_loader=test_loader,
            strategy=strategy
        )

        # Save first epoch timer
        if epoch == start_epoch:
            first_ep_t = timer()-lt

        # Final epoch
        if epoch + 1 == args.epochs:
            train_loader.last_epoch = True
            test_loader.last_epoch = True

        if strategy.is_main_worker:
            print('TIMER: epoch time:', timer()-lt, 's')
            print('DEBUG: accuracy:', acc_test, '%')

    if strategy.is_main_worker:
        print('\n--------------------------------------------------------')
        print('DEBUG: training results:\n')
        print('TIMER: first epoch time:', first_ep_t, ' s')
        print('TIMER: last epoch time:', timer()-lt, ' s')
        print('TIMER: average epoch time:', (timer()-et)/args.epochs, ' s')
        print('TIMER: total epoch time:', timer()-et, ' s')
        if epoch > 1:
            print('TIMER: total epoch-1 time:',
                  timer()-et-first_ep_t, ' s')
            print('TIMER: average epoch-1 time:',
                  (timer()-et-first_ep_t)/(args.epochs-1), ' s')
        print('DEBUG: last accuracy:', acc_test, '%')
        if torch.cuda.is_available():
            print('DEBUG: memory req:',
                  int(torch.cuda.memory_reserved(
                      strategy.local_rank())/1024/1024),
                  'MB')
            print('DEBUG: memory summary:\n\n',
                  torch.cuda.memory_summary(0))

        print(f'TIMER: final time: {timer()-st} s\n')

    time.sleep(1)
    print(f"<Global rank: {strategy.global_rank()}> - TRAINING FINISHED")

    # Clean-up
    strategy.clean_up()
    sys.exit()



"""
Framework-independent types.
"""


class MLArtifact:
    """A framework-independent machine learning artifact."""


class MLDataset(MLArtifact):
    """A framework-independent machine learning dataset."""


class MLModel(MLArtifact):
    """A framework-independent machine learning model."""



# This script trains the Convolutional Variational Auto-Encoder (CVAE)
# network on preprocessed CMIP6 Data


"""
from codecarbon import EmissionsTracker

# Instantiate the tracker object
tracker = EmissionsTracker(
    output_dir="../code_carbon/",  # define the directory where to write the emissions results
    output_file="emissions.csv",  # define the name of the file containing the emissions results
    # log_level='error' # comment out this line to see regular output
)
tracker.start()
"""

import torch
import torch.optim as optim
import numpy as np
import pandas as pd

from itwinai.components import Trainer, monitor_exec
import model
from torch.utils.data import DataLoader
from torchvision.utils import make_grid
from engine import train, validate
from utils import save_reconstructed_images, save_loss_plot, save_ex
from initialization import device, beta, criterion

class TorchTrainer(Trainer):
    def __init__(
            self,
            epochs: int,
            batch_size: int,
            lr: float
    ):
        super().__init__()
        self.epochs = epochs
        self.batch_size = batch_size
        self.lr = lr

    @monitor_exec
    def execute(self):
        # pick the season to study among:
        # '' (none, i.e. full dataset), 'winter_', 'spring_', 'summer_', 'autumn_'
        seasons = ["winter_", "spring_", "summer_", "autumn_"]

        # number of members used for the training of the network
        n_memb = 1

        # initialize learning parameters
        #lr0 = 0.001
        #batch_size = 64
        #epochs = 100
        #early stopping parameters
        stop_delta = 0.01  # under 1% improvement consider the model starts converging
        patience = 15  # wait for a few epochs to be sure before actually stopping
        early_count = 0  # count when validation loss < stop_delta
        old_valid_loss = 0  # keep track of validation loss at t-1

        for season in seasons:

            # initialize the model
            cvae_model = model.ConvVAE().to(device)
            optimizer = optim.Adam(cvae_model.parameters(), lr=self.lr)

            # load training set and train data
            train_time = pd.read_csv(f"input/dates_train_{season}data_{n_memb}memb.csv")
            train_data = np.load(
                f"input/preprocessed_1d_train_{season}data_{n_memb}memb.npy"
            )
            n_train = len(train_data)
            trainset = [
                (torch.from_numpy(np.reshape(train_data[i], (2, 32, 32))), train_time["0"][i])
                for i in range(n_train)
            ]
            # load train set, shuffle it, and create batches
            trainloader = DataLoader(trainset, batch_size=self.batch_size, shuffle=True)

            # load validation set and validation data
            test_time = pd.read_csv(f"input/dates_test_{season}data_{n_memb}memb.csv")
            test_data = np.load(f"input/preprocessed_1d_test_{season}data_{n_memb}memb.npy")
            n_test = len(test_data)
            testset = [
                (torch.from_numpy(np.reshape(test_data[i], (2, 32, 32))), test_time["0"][i])
                for i in range(n_test)
            ]
            testloader = DataLoader(testset, batch_size=self.batch_size, shuffle=False)

            # a list to save all the reconstructed images in PyTorch grid format
            grid_images = []
            # a list to save the loss evolutions
            train_loss = []
            valid_loss = []
            min_valid_epoch_loss = 100  # random high value

            for epoch in range(self.epochs):
                print(f"Epoch {epoch+1} of {self.epochs}")

                # train the model
                train_epoch_loss = train(
                    cvae_model, trainloader, trainset, device, optimizer, criterion, beta
                )

                # evaluate the model on the test set
                valid_epoch_loss, recon_images = validate(
                    cvae_model, testloader, testset, device, criterion, beta
                )

                # keep track of the losses
                train_loss.append(train_epoch_loss)
                valid_loss.append(valid_epoch_loss)

        # save the reconstructed images from the validation loop
        #save_reconstructed_images(recon_images, epoch+1, season)

        # convert the reconstructed images to PyTorch image grid format
        image_grid = make_grid(recon_images.detach().cpu())
        grid_images.append(image_grid)
        # save one example of reconstructed image before and after training

        #if epoch == 0 or epoch == self.epochs-1:
        #    save_ex(recon_images[0], epoch, season)

        # decreasing learning rate
        if (epoch + 1) % 20 == 0:
            lr = lr / 5

#-------


        # early stopping to avoid overfitting
#        if (
#            epoch > 1
#            and (old_valid_loss - valid_epoch_loss) / old_valid_loss < stop_delta
#        ):
            # if the marginal improvement in validation loss is too small
#            early_count += 1

        #if early_count > patience:
                # if too small improvement for a few epochs in a row, stop learning
        #        save_ex(recon_images[0], epoch, season)
                #break

#        else:
            # if the condition is not verified anymore, reset the count
#            early_count = 0
#        old_valid_loss = valid_epoch_loss

#---------------

        # save best model
        if valid_epoch_loss < min_valid_epoch_loss:
            min_valid_epoch_loss = valid_epoch_loss
            torch.save(
                cvae_model.state_dict(),
                f"outputs/cvae_model_{season}1d_{n_memb}memb.pth",
            )

            print(f"Train Loss: {train_epoch_loss:.4f}")
            print(f"Val Loss: {valid_epoch_loss:.4f}")

            save_loss_plot(train_loss, valid_loss, season)
            # save the loss evolutions
            pd.DataFrame(train_loss).to_csv(
                f"outputs/train_loss_indiv_{season}1d_{n_memb}memb.csv"
            )
            pd.DataFrame(valid_loss).to_csv(
                f"outputs/test_loss_indiv_{season}1d_{n_memb}memb.csv"
            )

        # emissions = tracker.stop()
        # print(f"Emissions from this training run: {emissions:.5f} kg CO2eq")



import tensorflow as tf

from ..scaling import minmax_transform


def get_tensor_decoding_fn(
    shape, drv_vars=[], coo_vars=None, msk_var=None,
    dtype=tf.float32
):

    def tensor_decoding_fn(serialized_data):
        """ Decoding function for a dataset written to disk as
        tensor_encoding_fn().
        """
        # define features dictionary
        features = {}

        # define all variables list
        vars = drv_vars.copy()
        if coo_vars:
            vars += coo_vars.copy()
        if msk_var:
            vars += [msk_var]

        # add driver + coordinate + mask vars to features
        for var in vars:
            features.update({var: tf.io.FixedLenFeature([], tf.string)})

        # parse the serialized data so we get a dict with our data.
        parsed_data = tf.io.parse_single_example(
            serialized_data, features=features)

        # accumulator for data elements
        data = []

        # get x raw data
        Xdrv = tf.stack([tf.ensure_shape(tf.io.parse_tensor(
            serialized=parsed_data[var], out_type=dtype),
            shape=shape)for var in drv_vars], axis=-1)
        data.append(Xdrv)

        # if coordinate vars are provided
        if coo_vars:
            if len(coo_vars) == 1:
                Ycoo = tf.ensure_shape(tf.io.parse_tensor(
                    serialized=parsed_data[coo_vars[0]], out_type=dtype),
                    shape=(2,))
            else:
                Ycoo = tf.stack([tf.ensure_shape(tf.io.parse_tensor(
                    serialized=parsed_data[var], out_type=dtype),
                    shape=(2)) for var in coo_vars], axis=-1)
            data.append(Ycoo)

        # if mask var is provided
        if msk_var:
            Ymsk = tf.expand_dims(tf.ensure_shape(tf.io.parse_tensor(
                serialized=parsed_data[msk_var], out_type=dtype),
                shape=shape), axis=-1)
            data.append(Ymsk)

        return tuple(data)

    return tensor_decoding_fn


def get_resize_fn(shape):

    def resize_fn(data):
        """Resize function that resizes the input data to the target shape."""
        resized_data = []
        for x in data:
            resized_data.append(tf.image.resize(
                x, shape, tf.image.ResizeMethod.NEAREST_NEIGHBOR))
        return tuple(resized_data)

    return resize_fn


def get_scaling_fn(scalers):

    def scaling_fn(data):
        """Function to scale the data according to the provided scalers."""
        data_scaled = []
        for x, scaler in zip(data, scalers):
            data_scaled.append(minmax_transform(x, scaler))
        return tuple(data_scaled)

    return scaling_fn


def get_scale_target_fn(label_no_cyclone, patch_size):

    def scale_target_fn(data):
        """Function to scale the target according to the provided
        label_no_cyclone"""
        x, y = data
        # scale y
        y_scaled = tf.math.divide(
            tf.subtract(tf.cast(y, dtype=tf.float32), label_no_cyclone),
            tf.subtract(tf.cast(patch_size-1, dtype=tf.float32),
                        label_no_cyclone)
        )
        return (x, y_scaled)

    return scale_target_fn


def get_masking_fn(mask):

    def masking_fn(data):
        # TODO : parametrize this function as the others
        X, y = data
        y_masked = tf.where(y < 0, mask, y)
        return (X, y_masked)

    return masking_fn


def read_tfrecord_as_tensor(filenames, shape, drv_vars, coo_vars, msk_var):
    # set autotune parameter to automatically manage resourches
    AUTOTUNE = tf.data.AUTOTUNE

    # get lambda functions to be applied to this dataset
    tensor_decoding_fn = get_tensor_decoding_fn(
        shape, drv_vars=drv_vars, coo_vars=coo_vars, msk_var=msk_var)

    # compute the number of samples into the dataset
    n_elems = sum(1 for _ in tf.data.TFRecordDataset(filenames))

    # Create standard dataset
    dataset = tf.data.TFRecordDataset(
        filenames, num_parallel_reads=AUTOTUNE).map(
        tensor_decoding_fn, num_parallel_calls=AUTOTUNE)

    # read data as numpy
    Xdata, ydata = dataset.batch(batch_size=n_elems).as_numpy_iterator().next()
    Xt = tf.convert_to_tensor(Xdata)
    yt = tf.convert_to_tensor(ydata)

    return Xt, yt



"""
In the first two tutorials we saw how to define simple sequential workflows by
means of the Pipeline object, which feds the outputs of the previous component
as inputs of the following one.
In this tutorial we show how to create more complex workflows, with
non-sequential data flows. Here, components can be arranges as an directed
acyclic graph (DAG). Under the DAG assumption, outputs of each block can be fed
as input potentially to any other component, granting great flexibility to the
experimenter.
The trade-off for improved flexibility is a change in the way we define
configuration files. From now on, it will only be possible to configure the
parameters used by the training script, but not its structure through the
Pipeline.

itwinai provides a wrapper of jsonarparse's ArgumentParser which supports
configuration files by default.

To run as usual:
>>> python my_script.py -d 20 --train-prop 0.7 --val-prop 0.2 --lr 1e-5

To reuse the parameters saved in a configuration file and override some
parameter (e.g., learning rate):
>>> python my_script.py --config advanced_tutorial_conf.yaml --lr 2e-3


"""
from typing import Any
from itwinai.parser import ArgumentParser
from itwinai.components import Predictor, monitor_exec

from basic_components import (
    MyDataGetter, MyDatasetSplitter, MyTrainer, MySaver
)


class MyEnsemblePredictor(Predictor):
    @monitor_exec
    def execute(self, dataset, model_ensemble) -> Any:
        """
        do some predictions with model on dataset...
        """
        return dataset


if __name__ == "__main__":
    parser = ArgumentParser(description="itwinai advanced workflows tutorial")
    parser.add_argument(
        "--data-size", "-d", type=int, required=True,
        help="Dataset cardinality.")
    parser.add_argument(
        "--train-prop", type=float, required=True,
        help="Train split proportion.")
    parser.add_argument(
        "--val-prop", type=float, required=True,
        help="Validation split proportion.")
    parser.add_argument(
        "--lr", type=float, help="Training learning rate.")
    args = parser.parse_args()

    # Save parsed arguments to configuration file.
    # Previous configurations are overwritten, which is not good,
    # but the versioning of configuration files is out of the scope
    # of this tutorial.
    parser.save(
        args, "advanced_tutorial_conf.yaml", format='yaml', overwrite=True)

    # Define workflow components
    getter = MyDataGetter(data_size=args.data_size)
    splitter = MyDatasetSplitter(
        train_proportion=args.train_prop,
        validation_proportion=args.val_prop,
        test_proportion=1-args.train_prop-args.val_prop
    )
    trainer1 = MyTrainer(lr=args.lr)
    trainer2 = MyTrainer(lr=args.lr)
    saver = MySaver()
    predictor = MyEnsemblePredictor(model=None)

    # Define ML workflow
    dataset = getter.execute()
    train_spl, val_spl, test_spl = splitter.execute(dataset)
    _, _, _, trained_model1 = trainer1.execute(train_spl, val_spl, test_spl)
    _, _, _, trained_model2 = trainer2.execute(train_spl, val_spl, test_spl)
    _ = saver.execute(trained_model1)
    predictions = predictor.execute(test_spl, [trained_model1, trained_model2])
    print()
    print("Predictions: " + str(predictions))



import torch
import torch.nn as nn
import torch.nn.functional as F

kernel_size = 4 # (4, 4) kernel
init_channels = 8 # initial number of filters
image_channels = 2 # 1 channel/variable: max temperature, precipitation, wind
latent_dim = 128 # latent space dimension (in which the image is compressed)


# define a Conv VAE
class ConvVAE(nn.Module):
    """
    A convolutional variational autoencoder:
    - encoder with 4 convolutional layers (ReLU activation functions)
    - latent space with 3 linear layers (one is double)
    - decoder with 4 deconvolutional layers (ReLUs, ending with a sigmoid)
    """

    def __init__(self):
        super(ConvVAE, self).__init__()

        # encoder
        self.enc1 = nn.Conv2d(
            in_channels=image_channels, out_channels=init_channels, kernel_size=kernel_size,
            stride=2, padding=1
        )
        self.enc2 = nn.Conv2d(
            in_channels=init_channels, out_channels=init_channels*2, kernel_size=kernel_size,
            stride=2, padding=1
        )
        self.enc3 = nn.Conv2d(
            in_channels=init_channels*2, out_channels=init_channels*4, kernel_size=kernel_size,
            stride=2, padding=1
        )
        self.enc4 = nn.Conv2d(
            in_channels=init_channels*4, out_channels=64, kernel_size=kernel_size,
            stride=2, padding=0
        )

        # fully connected layers for learning representations
        self.fc1 = nn.Linear(64, 128)
        self.fc_mu = nn.Linear(128, latent_dim)
        self.fc_log_var = nn.Linear(128, latent_dim)
        self.fc2 = nn.Linear(latent_dim, 64)

        # decoder
        self.dec1 = nn.ConvTranspose2d(
            in_channels=64, out_channels=init_channels*8, kernel_size=kernel_size,
            stride=1, padding=0
        )
        self.dec2 = nn.ConvTranspose2d(
            in_channels=init_channels*8, out_channels=init_channels*4, kernel_size=kernel_size,
            stride=2, padding=1
        )
        self.dec3 = nn.ConvTranspose2d(
            in_channels=init_channels*4, out_channels=init_channels*2, kernel_size=kernel_size,
            stride=2, padding=1
        )
        self.dec4 = nn.ConvTranspose2d(
            in_channels=init_channels*2, out_channels=image_channels, kernel_size=kernel_size,
            stride=2, padding=1
        )

    def reparameterize(self, mu, log_var):
        """
        Reparametrization trick.
        :param mu: mean from the encoder's latent space
        :param log_var: log variance from the encoder's latent space
        """
        std = torch.exp(0.5*log_var) # standard deviation
        eps = torch.randn_like(std) # `randn_like` as we need the same size
        sample = mu + (eps * std) # sampling
        return sample

    def forward(self, x):
        # encoding
        x = F.relu(self.enc1(x))
        x = F.relu(self.enc2(x))
        x = F.relu(self.enc3(x))
        x = F.relu(self.enc4(x))
        batch, _, _, _ = x.shape
        x = F.adaptive_avg_pool2d(x, 1).reshape(batch, -1)
        hidden = self.fc1(x)

        # get `mu` and `log_var`
        mu = self.fc_mu(hidden)
        log_var = self.fc_log_var(hidden)
        # get the latent vector through reparameterization
        z = self.reparameterize(mu, log_var)
        z = self.fc2(z)
        z = z.view(-1, 64, 1, 1)

        # decoding
        x = F.relu(self.dec1(z))
        x = F.relu(self.dec2(x))
        x = F.relu(self.dec3(x))
        reconstruction = torch.sigmoid(self.dec4(x))
        return reconstruction, mu, log_var



# import tensorflow.keras as keras
import tensorflow as tf

from typing import List


class MNIST_Model(tf.keras.Model):
    def __init__(
        self,
        input_shape: List[int] = (28, 28, 1),
        output_shape: int = 10
    ):
        super().__init__()

        self.model = tf.keras.Sequential(
            [
                # tf.keras.Input(shape=input_shape),
                tf.keras.layers.Conv2D(
                    32, kernel_size=3, activation="relu",
                    input_shape=input_shape),
                tf.keras.layers.MaxPooling2D(2),
                # tf.keras.layers.Conv2D(
                #     64, kernel_size=3, activation="relu"),
                # tf.keras.layers.MaxPooling2D(pool_size=2),
                tf.keras.layers.Flatten(),
                # tf.keras.layers.Dropout(0.5),
                tf.keras.layers.Dense(output_shape)
            ]
        )

        self.model = tf.keras.Sequential([
            tf.keras.layers.Conv2D(
                32, 3, activation='relu', input_shape=input_shape),
            tf.keras.layers.MaxPooling2D(2),
            tf.keras.layers.Flatten(),
            tf.keras.layers.Dense(64, activation='relu'),
            tf.keras.layers.Dense(output_shape)
        ])

        # LeNet5
        self.model = tf.keras.Sequential([

            tf.keras.layers.Conv2D(filters=6, kernel_size=(
                3, 3), activation='relu', input_shape=(28, 28, 1)),
            tf.keras.layers.AveragePooling2D(2),
            tf.keras.layers.Conv2D(
                filters=16, kernel_size=(3, 3), activation='relu'),
            tf.keras.layers.AveragePooling2D(2),
            tf.keras.layers.Flatten(),
            tf.keras.layers.Dense(units=120, activation='relu'),
            tf.keras.layers.Dense(units=84, activation='relu'),
            tf.keras.layers.Dense(units=10)
        ])

    def call(self, inputs):
        return self.model(inputs)



"""
Utilities for itwinai package.
"""
import os
import yaml

from collections.abc import MutableMapping
from typing import Dict
from omegaconf import OmegaConf
from omegaconf.dictconfig import DictConfig


def load_yaml(path: str) -> Dict:
    """Load YAML file as dict.

    Args:
        path (str): path to YAML file.

    Raises:
        exc: yaml.YAMLError for loading/parsing errors.

    Returns:
        Dict: nested dict representation of parsed YAML file.
    """
    with open(path, "r", encoding="utf-8") as yaml_file:
        try:
            loaded_config = yaml.safe_load(yaml_file)
        except yaml.YAMLError as exc:
            print(exc)
            raise exc
    return loaded_config


def load_yaml_with_deps_from_file(path: str) -> DictConfig:
    """
    Load YAML file with OmegaConf and merge it with its dependencies
    specified in the `conf-dependencies` field.
    Assume that the dependencies live in the same folder of the
    YAML file which is importing them.

    Args:
        path (str): path to YAML file.

    Raises:
        exc: yaml.YAMLError for loading/parsing errors.

    Returns:
        DictConfig: nested representation of parsed YAML file.
    """
    yaml_conf = load_yaml(path)
    use_case_dir = os.path.dirname(path)
    deps = []
    if yaml_conf.get("conf-dependencies"):
        for dependency in yaml_conf["conf-dependencies"]:
            deps.append(load_yaml(os.path.join(use_case_dir, dependency)))

    return OmegaConf.merge(yaml_conf, *deps)


def load_yaml_with_deps_from_dict(dict_conf, use_case_dir) -> DictConfig:
    deps = []

    if dict_conf.get("conf-dependencies"):
        for dependency in dict_conf["conf-dependencies"]:
            deps.append(load_yaml(os.path.join(use_case_dir, dependency)))

    return OmegaConf.merge(dict_conf, *deps)


def dynamically_import_class(name: str):
    """
    Dynamically import class by module path.
    Adapted from https://stackoverflow.com/a/547867

    Args:
        name (str): path to the class (e.g., mypackage.mymodule.MyClass)

    Returns:
        __class__: class object.
    """
    module, class_name = name.rsplit(".", 1)
    mod = __import__(module, fromlist=[class_name])
    klass = getattr(mod, class_name)
    return klass


def flatten_dict(
    d: MutableMapping, parent_key: str = "", sep: str = "."
) -> MutableMapping:
    """Flatten dictionary

    Args:
        d (MutableMapping): nested dictionary to flatten
        parent_key (str, optional): prefix for all keys. Defaults to ''.
        sep (str, optional): separator for nested key concatenation.
            Defaults to '.'.

    Returns:
        MutableMapping: flattened dictionary with new keys.
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)



import pytest

pytest.PIPE_LIST_YAML = """
my-list-pipeline:
  class_path: itwinai.pipeline.Pipeline
  init_args:
    steps:
      - class_path: itwinai.tests.dummy_components.FakePreproc
        init_args:
          max_items: 32
          name: my-preproc

      - class_path: itwinai.tests.dummy_components.FakeTrainer
        init_args:
          lr: 0.001
          batch_size: 32
          name: my-trainer

      - class_path: itwinai.tests.dummy_components.FakeSaver
        init_args:
          save_path: ./some/path
          name: my-saver
"""

pytest.PIPE_DICT_YAML = """
my-dict-pipeline:
  class_path: itwinai.pipeline.Pipeline
  init_args:
    steps:
      preproc-step:
        class_path: itwinai.tests.dummy_components.FakePreproc
        init_args:
          max_items: 32
          name: my-preproc

      train-step:
        class_path: itwinai.tests.dummy_components.FakeTrainer
        init_args:
          lr: 0.001
          batch_size: 32
          name: my-trainer

      save-step:
        class_path: itwinai.tests.dummy_components.FakeSaver
        init_args:
          save_path: ./some/path
          name: my-saver
"""

pytest.NESTED_PIPELINE = """
some:
  field:
    nst-pipeline:
      class_path: itwinai.pipeline.Pipeline
      init_args:
        steps:
          - class_path: itwinai.tests.FakePreproc
            init_args:
              max_items: 32
              name: my-preproc

          - class_path: itwinai.tests.FakeTrainer
            init_args:
              lr: 0.001
              batch_size: 32
              name: my-trainer

          - class_path: itwinai.tests.FakeSaver
            init_args:
              save_path: ./some/path
              name: my-saver
"""



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys
import subprocess

exclude_patterns = 'requirements.txt'

# sys.path.insert(0, os.path.abspath('../use-cases/'))
# sys.path.insert(0, os.path.abspath('../use-cases/3dgan/'))
# sys.path.insert(0, os.path.abspath('../use-cases/mnist/torch-lightning/'))
# sys.path.insert(0, os.path.abspath('../use-cases/mnist/torch/'))
sys.path.insert(0, os.path.abspath('../tutorials/ml-workflows/'))
# sys.path.insert(0, os.path.abspath('../tutorials/distributed-ml/'))
# sys.path.insert(0, os.path.abspath('../src/itwinai'))
# sys.path.insert(0, os.path.abspath('../src/itwinai/tensorflow'))
# sys.path.insert(0, os.path.abspath('../src/itwinai/torch'))
sys.path.insert(0, os.path.abspath('../src'))
# sys.path.insert(0, os.path.abspath('../...'))

project = 'itwinai'
copyright = ('2024, Matteo Bunino, Kalliopi Tsolaki, '
             'Rakesh Sarma, Mario Ruettgers on behalf of CERN & JSC')
author = 'Matteo Bunino, Kalliopi Tsolaki, Rakesh Sarma, Mario Ruettgers'
version = '0.0'  # short version
release = '0.0.2'  # full version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.autodoc', 'sphinx.ext.doctest',
              'sphinx.ext.viewcode', 'myst_parser', 'nbsphinx',
              'sphinx.ext.napoleon']

source_suffix = {
    '.rst': 'restructuredtext',
    '.txt': 'markdown',
    '.md': 'markdown'  # ,
    # '.ipynb': 'nbsphinx'
}

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

autodoc_mock_imports = ["mlflow"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output


def get_git_tag():
    try:
        return subprocess.check_output(
            ['git', 'describe', '--tags', '--abbrev=0']
        ).decode('utf-8').strip()
    except subprocess.CalledProcessError:
        return 'unknown'


# Set the version to the latest tag
version = get_git_tag()
release = version

html_theme = 'sphinx_rtd_theme'  # 'alabaster'
html_static_path = ['_static']

html_context = {
    'display_version': True,
    'release': release
}

html_footer = """
<div class="custom-footer">
    <strong>Version:</strong> {{ release }}
</div>
"""

html_sidebars = {
    '**': [
        html_footer  # Adds the custom footer with version information
    ]
}



"""Abstraction layer for loggers."""
import os
import csv
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from typing import Any, Dict, List, Optional, Union, Literal
import pickle
import pathlib

import wandb
import mlflow

BASE_EXP_NAME: str = 'default_experiment'


class LogMixin(metaclass=ABCMeta):
    @abstractmethod
    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = 'metric',
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        **kwargs
    ) -> None:
        """Log ``item`` with ``identifier`` name of ``kind`` type at ``step``
        time step.

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be one
                among the list of self.supported_types. Defaults to 'metric'.
            step (Optional[int], optional): logging step. Defaults to None.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
        """


class Logger(LogMixin, metaclass=ABCMeta):
    """Base class for logger

    Args:
        savedir (str, optional): filesystem location where logs are stored.
            Defaults to 'mllogs'.
        log_freq (Union[int, Literal['epoch', 'batch']], optional):
            how often should the logger fulfill calls to the `log()`
            method:

            - When set to 'epoch', the logger logs only if ``batch_idx``
              is not passed to the ``log`` method.

            - When an integer
              is given, the logger logs if ``batch_idx`` is a multiple of
              ``log_freq``.

            - When set to ``'batch'``, the logger logs always.

            Defaults to 'epoch'.
    """
    #: Location on filesystem where to store data.
    savedir: str = None
    #: Supported logging 'kind's.
    supported_types: List[str]
    _log_freq: Union[int, Literal['epoch', 'batch']]

    def __init__(
        self,
        savedir: str = 'mllogs',
        log_freq: Union[int, Literal['epoch', 'batch']] = 'epoch'
    ) -> None:
        self.savedir = savedir
        self.log_freq = log_freq

    @property
    def log_freq(self) -> Union[int, Literal['epoch', 'batch']]:
        """Get ``log_feq``, namely how often should the logger
        fulfill or ignore calls to the `log()` method."""
        return self._log_freq

    @log_freq.setter
    def log_freq(self, val: Union[int, Literal['epoch', 'batch']]):
        """Sanitize log_freq value."""
        if val in ['epoch', 'batch'] or (isinstance(val, int) and val > 0):
            self._log_freq = val
        else:
            raise ValueError(
                "Wrong value for 'log_freq'. Supported values are: "
                f"['epoch', 'batch'] or int > 0. Received: {val}"
            )

    @contextmanager
    def start_logging(self):
        """Start logging context.

        Example:


        >>> with my_logger.start_logging():
        >>>     my_logger.log(123, 'value', kind='metric', step=0)


        """
        try:
            self.create_logger_context()
            yield
        finally:
            self.destroy_logger_context()

    @abstractmethod
    def create_logger_context(self):
        """Initialize logger."""

    @abstractmethod
    def destroy_logger_context(self):
        """Destroy logger."""

    @abstractmethod
    def save_hyperparameters(self, params: Dict[str, Any]) -> None:
        """Save hyperparameters.

        Args:
            params (Dict[str, Any]): hyperparameters dictionary.
        """

    def serialize(self, obj: Any, identifier: str) -> str:
        """Serializes object to disk and returns its path.

        Args:
            obj (Any): item to save.
            identifier (str): identifier of the item to log (expected to be a
                path under ``self.savedir``).

        Returns:
            str: local path of the serialized object to be logged.
        """
        itm_path = os.path.join(self.savedir, identifier)
        with open(itm_path, 'wb') as itm_file:
            pickle.dump(obj, itm_file)

    def should_log(
        self,
        batch_idx: Optional[int] = None
    ) -> bool:
        """Determines whether the logger should fulfill or ignore calls to the
        `log()` method, depending on the ``log_freq`` property:

        - When ``log_freq`` is set to 'epoch', the logger logs only if
          ``batch_idx`` is not passed to the ``log`` method.

        - When ``log_freq`` is an integer
          is given, the logger logs if ``batch_idx`` is a multiple of
          ``log_freq``.

        - When ``log_freq`` is set to ``'batch'``, the logger logs always.

        Args:
            batch_idx (Optional[int]): the dataloader batch idx, if available.
                Defaults to None.

        Returns:
            bool: True if the logger should log, False otherwise.
        """
        if batch_idx is not None:
            if isinstance(self.log_freq, int):
                if batch_idx % self.log_freq == 0:
                    return True
                return False
            if self.log_freq == 'batch':
                return True
            return False
        return True


class ConsoleLogger(Logger):
    """Simplified logger.

    Args:
        savedir (str, optional): where to store artifacts.
            Defaults to 'mllogs'.
        log_freq (Union[int, Literal['epoch', 'batch']], optional):
            determines whether the logger should fulfill or ignore
            calls to the `log()` method. See ``Logger.should_log`` method for
            more details. Defaults to 'epoch'.
    """

    def __init__(
        self,
        savedir: str = 'mllogs',
        log_freq: Union[int, Literal['epoch', 'batch']] = 'epoch'
    ) -> None:
        savedir = os.path.join(savedir, 'simple-logger')
        super().__init__(savedir=savedir, log_freq=log_freq)
        self.supported_types = ['torch', 'artifact']

    def create_logger_context(self):
        """Initialize logger."""
        os.makedirs(self.savedir, exist_ok=True)
        run_dirs = sorted([int(dir) for dir in os.listdir(self.savedir)])
        if len(run_dirs) == 0:
            self.run_id = 0
        else:
            self.run_id = int(run_dirs[-1]) + 1
        self.run_path = os.path.join(self.savedir, str(self.run_id))
        os.makedirs(self.run_path)

    def destroy_logger_context(self):
        """Destroy logger. Do nothing."""

    def save_hyperparameters(self, params: Dict[str, Any]) -> None:
        """Save hyperparameters. Do nothing.

        Args:
            params (Dict[str, Any]): hyperparameters dictionary.
        """

    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = 'metric',
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        **kwargs
    ) -> None:
        """Print metrics to stdout and save artifacts to the filesystem.

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be
                one among the list of ``self.supported_types``.
                Defaults to 'metric'.
            step (Optional[int], optional): logging step. Defaults to None.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
            kwargs: keyword arguments to pass to the logger.
        """
        if not self.should_log(batch_idx=batch_idx):
            return

        if kind == 'artifact':
            if isinstance(item, str) and os.path.isfile(item):
                import shutil
                identifier = os.path.join(
                    self.run_path,
                    identifier
                )
                print(f"ConsoleLogger: Serializing to {identifier}...")
                shutil.copyfile(item, identifier)
            else:
                identifier = os.path.join(
                    os.path.basename(self.run_path),
                    identifier
                )
                print(f"ConsoleLogger: Serializing to {identifier}...")
                self.serialize(item, identifier)
        elif kind == 'torch':
            identifier = os.path.join(self.run_path, identifier)
            print(f"ConsoleLogger: Saving to {identifier}...")
            import torch
            torch.save(item, identifier)
        else:
            print(f"ConsoleLogger: {identifier} = {item}")


class MLFlowLogger(Logger):
    """Abstraction around MLFlow logger.

    Args:
        savedir (str, optional): path on local filesystem where logs are
            stored. Defaults to 'mllogs'.
        experiment_name (str, optional): experiment name. Defaults to
            ``itwinai.loggers.BASE_EXP_NAME``.
        tracking_uri (Optional[str], optional): MLFLow tracking URI.
            Overrides ``savedir`` if given. Defaults to None.
        run_description (Optional[str], optional): run description.
            Defaults to None.
        log_freq (Union[int, Literal['epoch', 'batch']], optional):
            determines whether the logger should fulfill or ignore
            calls to the `log()` method. See ``Logger.should_log`` method for
            more details. Defaults to 'epoch'.
    """

    #: Current MLFLow experiment's run.
    active_run: mlflow.ActiveRun

    def __init__(
        self,
        savedir: str = 'mllogs',
        experiment_name: str = BASE_EXP_NAME,
        tracking_uri: Optional[str] = None,
        run_description: Optional[str] = None,
        log_freq: Union[int, Literal['epoch', 'batch']] = 'epoch'
    ):
        savedir = os.path.join(savedir, 'mlflow')
        super().__init__(savedir=savedir, log_freq=log_freq)
        self.experiment_name = experiment_name
        self.tracking_uri = tracking_uri
        self.run_description = run_description

        if self.tracking_uri is None:
            # Default MLFLow tracking URI
            saved_abs_path = os.path.abspath(self.savedir)
            self.tracking_uri = pathlib.Path(saved_abs_path).as_uri()

        # TODO: for pytorch lightning:
        # mlflow.pytorch.autolog()

        self.supported_types = [
            'metric', 'figure', 'image', 'artifact', 'torch', 'dict', 'param',
            'text'
        ]

    def create_logger_context(self):
        """Initialize logger. Start MLFLow run."""
        mlflow.set_tracking_uri(self.tracking_uri)
        mlflow.set_experiment(experiment_name=self.experiment_name)
        self.active_run: mlflow.ActiveRun = mlflow.start_run(
            description=self.run_description
        )

    def destroy_logger_context(self):
        """Destroy logger. End current MLFlow run."""
        mlflow.end_run()

    def save_hyperparameters(self, params: Dict[str, Any]) -> None:
        """Save hyperparameters as MLFlow parameters.

        Args:
            params (Dict[str, Any]): hyperparameters dictionary.
        """
        for param_name, val in params.items():
            self.log(item=val, identifier=param_name, step=0, kind='param')

    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = 'metric',
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        **kwargs
    ) -> None:
        """Log with MLFlow.

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be
                one among the list of ``self.supported_types``.
                Defaults to 'metric'.
            step (Optional[int], optional): logging step. Defaults to None.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
            kwargs: keyword arguments to pass to the logger.
        """
        if not self.should_log(batch_idx=batch_idx):
            return

        if kind == 'metric':
            # if isinstance(item, list) and isinstance(identifier, list):
            mlflow.log_metric(
                key=identifier,
                value=item,
                step=step
            )
        if kind == 'artifact':
            if not isinstance(item, str):
                # Save the object locally and then log it
                name = os.path.basename(identifier)
                save_path = os.path.join(self.savedir, '.trash', name)
                os.makedirs(os.path.dirname(save_path), exist_ok=True)
                item = self.serialize(item, save_path)
            mlflow.log_artifact(
                local_path=item,
                artifact_path=identifier
            )
        if kind == 'torch':
            import torch
            # Save the object locally and then log it
            name = os.path.basename(identifier)
            save_path = os.path.join(self.savedir, '.trash', name)
            os.makedirs(os.path.dirname(save_path), exist_ok=True)
            torch.save(item, save_path)
            # Log into mlflow
            mlflow.log_artifact(
                local_path=save_path,
                artifact_path=identifier
            )
        if kind == 'dict':
            mlflow.log_dict(
                dictionary=item,
                artifact_file=identifier
            )
        if kind == 'figure':
            mlflow.log_figure(
                artifact_file=identifier,
                figure=item,
                save_kwargs=kwargs.get('save_kwargs')
            )
        if kind == 'image':
            mlflow.log_image(
                artifact_file=identifier,
                image=item
            )
        if kind == 'param':
            mlflow.log_param(
                key=identifier,
                value=item
            )
        if kind == 'text':
            mlflow.log_text(
                artifact_file=identifier,
                text=item
            )


class WandBLogger(Logger):
    """Abstraction around WandB logger.

    Args:
        savedir (str, optional): location on local filesystem where logs
            are stored. Defaults to 'mllogs'.
        project_name (str, optional): experiment name. Defaults to
            ``itwinai.loggers.BASE_EXP_NAME``.
        log_freq (Union[int, Literal['epoch', 'batch']], optional):
            determines whether the logger should fulfill or ignore
            calls to the `log()` method. See ``Logger.should_log`` method for
            more details. Defaults to 'epoch'.
    """

    # TODO: add support for artifacts logging

    def __init__(
        self,
        savedir: str = 'mllogs',
        project_name: str = BASE_EXP_NAME,
        log_freq: Union[int, Literal['epoch', 'batch']] = 'epoch'
    ) -> None:
        savedir = os.path.join(savedir, 'wandb')
        super().__init__(savedir=savedir, log_freq=log_freq)
        self.project_name = project_name
        self.supported_types = [
            'watch', 'metric', 'figure', 'image', 'torch', 'dict',
            'param', 'text'
        ]

    def create_logger_context(self):
        """Initialize logger. Init WandB run."""
        os.makedirs(os.path.join(self.savedir, 'wandb'), exist_ok=True)
        self.active_run = wandb.init(
            dir=os.path.abspath(self.savedir),
            project=self.project_name
        )

    def destroy_logger_context(self):
        """Destroy logger."""

    def save_hyperparameters(self, params: Dict[str, Any]) -> None:
        """Save hyperparameters.

        Args:
            params (Dict[str, Any]): hyperparameters dictionary.
        """
        wandb.config.update(params)

    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = 'metric',
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        **kwargs
    ) -> None:
        """Log with WandB. Wrapper of https://docs.wandb.ai/ref/python/log

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be
                one among the list of ``self.supported_types``.
                Defaults to 'metric'.
            step (Optional[int], optional): ignored by ``WandBLogger``.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
            kwargs: keyword arguments to pass to the logger.
        """
        if not self.should_log(batch_idx=batch_idx):
            return

        if kind == 'watch':
            wandb.watch(item)
        if kind in self.supported_types[1:]:
            # wandb.log({identifier: item}, step=step, commit=True)
            # Let WandB use its preferred step
            wandb.log({identifier: item}, commit=True)


class TensorBoardLogger(Logger):
    """Abstraction around TensorBoard logger, both for PyTorch and
    TensorFlow.

    Args:
        savedir (str, optional): location on local filesystem where logs
            are stored. Defaults to 'mllogs'.
        log_freq (Union[int, Literal['epoch', 'batch']], optional):
            determines whether the logger should fulfill or ignore
            calls to the `log()` method. See ``Logger.should_log`` method for
            more details. Defaults to 'epoch'.
        framework (Literal['tensorflow', 'pytorch'], optional):
            whether to log PyTorch or TensorFlow ML data.
            Defaults to 'pytorch'.

    Raises:
        ValueError: when ``framework`` is not recognized.
    """

    # TODO: decouple the logger into TorchTBLogger and TFTBLogger
    # and add the missing logging types supported by each.

    def __init__(
        self,
        savedir: str = 'mllogs',
        log_freq: Union[int, Literal['epoch', 'batch']] = 'epoch',
        framework: Literal['tensorflow', 'pytorch'] = 'pytorch'
    ) -> None:
        savedir = os.path.join(savedir, 'tensorboard')
        super().__init__(savedir=savedir, log_freq=log_freq)
        self.framework = framework
        if framework.lower() == 'tensorflow':
            import tensorflow as tf
            self.tf = tf
            self.writer = tf.summary.create_file_writer(savedir)
        elif framework.lower() == 'pytorch':
            from torch.utils.tensorboard import SummaryWriter
            self.writer = SummaryWriter(savedir)
        else:
            raise ValueError(
                "Framework must be either 'tensorflow' or 'pytorch'")
        self.supported_types = ['metric', 'image',
                                'text', 'figure', 'torch']

    def create_logger_context(self):
        """Initialize logger."""
        if self.framework == 'tensorflow':
            self.writer.set_as_default()

    def destroy_logger_context(self):
        """Destroy logger. Close SummaryWriter."""
        self.writer.close()

    def save_hyperparameters(self, params: Dict[str, Any]) -> None:
        """Save hyperparameters.

        Args:
            params (Dict[str, Any]): hyperparameters dictionary.
        """
        if self.framework == 'tensorflow':
            from tensorboard.plugins.hparams import api as hp
            hparams = {hp.HParam(k): v for k, v in params.items()}
            with self.writer.as_default():
                hp.hparams(hparams)
        elif self.framework == 'pytorch':
            self.writer.add_hparams(params, {})

    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = 'metric',
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        **kwargs
    ) -> None:
        """Log with Tensorboard.

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be
                one among the list of ``self.supported_types``.
                Defaults to 'metric'.
            step (Optional[int], optional): logging step. Defaults to None.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
            kwargs: keyword arguments to pass to the logger.
        """
        if not self.should_log(batch_idx=batch_idx):
            return

        if self.framework == 'tensorflow':
            with self.writer.as_default():
                if kind == 'metric':
                    self.tf.summary.scalar(identifier, item, step=step)
                elif kind == 'image':
                    self.tf.summary.image(identifier, item, step=step)
                elif kind == 'text':
                    self.tf.summary.text(identifier, item, step=step)
                elif kind == 'figure':
                    self.tf.summary.figure(identifier, item, step=step)
        elif self.framework == 'pytorch':
            if kind == 'metric':
                self.writer.add_scalar(identifier, item, global_step=step)
            elif kind == 'image':
                self.writer.add_image(identifier, item, global_step=step)
            elif kind == 'text':
                self.writer.add_text(identifier, item, global_step=step)
            elif kind == 'figure':
                self.writer.add_figure(identifier, item, global_step=step)
            elif kind == 'torch':
                self.writer.add_graph(item)


class LoggersCollection(Logger):
    """Wrapper of a set of loggers, allowing to use them simultaneously.

    Args:
        loggers (List[Logger]): list of itwinai loggers.
    """

    def __init__(
        self,
        loggers: List[Logger]
    ) -> None:
        super().__init__(savedir='/tmp/mllogs_LoggersCollection', log_freq=1)
        self.loggers = loggers

    def should_log(self, batch_idx: int = None) -> bool:
        """Transparent method which delegates the `Logger.should_log``
        to individual loggers. Always returns True.

        Args:
            batch_idx (int, optional): dataloader batch index.
            Defaults to None.

        Returns:
            bool: always True.
        """
        return True

    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = 'metric',
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        **kwargs
    ) -> None:
        """Log on all loggers.

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be
                one among the list of ``self.supported_types``.
                Defaults to 'metric'.
            step (Optional[int], optional): logging step. Defaults to None.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
            kwargs: keyword arguments to pass to the logger.
        """
        for logger in self.loggers:
            logger.log(
                item=item,
                identifier=identifier,
                kind=kind,
                step=step,
                batch_idx=batch_idx,
                **kwargs
            )

    def create_logger_context(self):
        """Initialize all loggers."""
        for logger in self.loggers:
            logger.create_logger_context()

    def destroy_logger_context(self):
        """Destroy all loggers."""
        for logger in self.loggers:
            logger.destroy_logger_context()

    def save_hyperparameters(self, params: Dict[str, Any]) -> None:
        """Save hyperparameters for all loggers.

        Args:
            params (Dict[str, Any]): hyperparameters dictionary.
        """
        for logger in self.loggers:
            logger.save_hyperparameters(params=params)


class EpochTimeTracker:
    """Profiler for epoch execution time used to support scaling tests.
    It uses CSV files to store, for each epoch, the ``name`` of the
    experiment, the number of compute ``nodes`` used, the ``epoch_id``,
    and the execution ``time`` in seconds.

    Args:
        series_name (str): name of the experiment/job.
        csv_file (str): path to CSV file to store experiments times.
    """

    def __init__(self, series_name: str, csv_file: str) -> None:
        self.series_name = series_name
        self._data = []
        self.csv_file = csv_file
        with open(csv_file, 'w') as csvfile:
            csvwriter = csv.writer(csvfile)
            csvwriter.writerow(['name', 'nodes', 'epoch_id', 'time'])

    def add_epoch_time(self, epoch_idx: int, time: float) -> None:
        """Add row to the current experiment's CSV file in append mode.

        Args:
            epoch_idx (int): epoch order idx.
            time (float): epoch execution time (seconds).
        """
        n_nodes = os.environ.get('SLURM_NNODES', -1)
        fields = (self.series_name, n_nodes, epoch_idx, time)
        self._data.append(fields)
        with open(self.csv_file, 'a') as csvfile:
            csvwriter = csv.writer(csvfile)
            csvwriter.writerow(fields)

    def save(self, csv_file: Optional[str] = None) -> None:
        """Save data to a new CSV file.

        Args:
            csv_file (Optional[str], optional): path to the CSV file.
                If not given, uses the one given in the constructor.
                Defaults to None.
        """
        if not csv_file:
            csv_file = self.csv_file
        with open(csv_file, 'w') as csvfile:
            csvwriter = csv.writer(csvfile)
            csvwriter.writerow(['name', 'nodes', 'epoch_id', 'time'])
            csvwriter.writerows(self._data)



from typing import Dict, Any, Union
import abc
import json
import yaml
from pathlib import Path
import inspect

from .type import MLModel
from .utils import SignatureInspector


def is_jsonable(x):
    try:
        json.dumps(x)
        return True
    except Exception:
        return False


def fullname(o):
    klass = o.__class__
    module = klass.__module__
    if module == 'builtins':
        return klass.__qualname__  # avoid outputs like 'builtins.str'
    return module + '.' + klass.__qualname__


class SerializationError(Exception):
    ...


class Serializable:
    #: Dictionary storing constructor arguments. Needed to serialize the
    #: class to dictionary. Set by ``self.save_parameters()`` method.
    parameters: Dict[Any, Any] = None

    def save_parameters(self, **kwargs) -> None:
        """Simplified way to store constructor arguments in as class
        attributes. Keeps track of the parameters to enable
        YAML/JSON serialization.
        """
        if self.parameters is None:
            self.parameters = {}
        self.parameters.update(kwargs)

        # for k, v in kwargs.items():
        #     self.__setattr__(k, v)

    @staticmethod
    def locals2params(locals: Dict, pop_self: bool = True) -> Dict:
        """Remove ``self`` from the output of ``locals()``.

        Args:
            locals (Dict): output of ``locals()`` called in the constructor
                of a class.
            pop_self (bool, optional): whether to remove ``self``.
                Defaults to True.

        Returns:
            Dict: cleaned ``locals()``.
        """
        if pop_self:
            locals.pop('self', None)
        return locals

    def update_parameters(self, **kwargs) -> None:
        """Updates stored parameters."""
        self.save_parameters(**kwargs)

    def to_dict(self) -> Dict:
        """Returns a dict serialization of the current object."""
        self._validate_parameters()
        class_path = self._get_class_path()
        init_args = dict()
        for par_name, par in self._saved_constructor_parameters().items():
            init_args[par_name] = self._recursive_serialization(par, par_name)
        return dict(class_path=class_path, init_args=init_args)

    def _validate_parameters(self) -> None:
        if self.parameters is None:
            raise SerializationError(
                f"{self.__class__.__name__} cannot be serialized "
                "because its constructor arguments were not saved. "
                "Please add 'self.save_parameters(param_1=param_1, "
                "..., param_n=param_n)' as first instruction of its "
                "constructor."
            )

        init_inspector = SignatureInspector(self.__init__)
        for par_name in init_inspector.required_params:
            if self.parameters.get(par_name) is None:
                raise SerializationError(
                    f"Required parameter '{par_name}' of "
                    f"{self.__class__.__name__} class not present in "
                    "saved parameters. "
                    "Please add 'self.save_parameters(param_1=param_1, "
                    "..., param_n=param_n)' as first instruction of its "
                    f"constructor, including also '{par_name}'."
                )

    def _get_class_path(self) -> str:
        class_path = fullname(self)
        if "<locals>" in class_path:
            raise SerializationError(
                f"{self.__class__.__name__} is "
                "defined locally, which is not supported for serialization. "
                "Move the class to a separate Python file and import it "
                "from there."
            )
        return class_path

    def _saved_constructor_parameters(self) -> Dict[str, Any]:
        """Extracts the current constructor parameters from all
        the saved parameters, as some of them may had been added by
        superclasses.

        Returns:
            Dict[str, Any]: subset of saved parameters containing only
            the constructor parameters for this class.
        """
        init_params = inspect.signature(self.__init__).parameters.items()
        init_par_nam = map(lambda x: x[0], init_params)
        return {
            par_name: self.parameters[par_name] for par_name in init_par_nam
            if self.parameters.get(par_name, inspect._empty) != inspect._empty
        }

    def _recursive_serialization(self, item: Any, item_name: str) -> Any:
        if isinstance(item, (tuple, list, set)):
            return [self._recursive_serialization(x, item_name) for x in item]
        elif isinstance(item, dict):
            return {
                k: self._recursive_serialization(v, item_name)
                for k, v in item.items()
            }
        elif is_jsonable(item):
            return item
        elif isinstance(item, Serializable):
            return item.to_dict()
        else:
            raise SerializationError(
                f"{self.__class__.__name__} cannot be serialized "
                f"because its constructor argument '{item_name}' "
                "is not a Python built-in type and it does not "
                "extend 'itwinai.serialization.Serializable' class."
            )

    def to_json(self, file_path: Union[str, Path], nested_key: str) -> None:
        """Save a component to JSON file.

        Args:
            file_path (Union[str, Path]): JSON file path.
            nested_key (str): root field containing the serialized object.
        """
        with open(file_path, "w") as fp:
            json.dump({nested_key: self.to_dict()}, fp)

    def to_yaml(self, file_path: Union[str, Path], nested_key: str) -> None:
        """Save a component to YAML file.

        Args:
            file_path (Union[str, Path]): YAML file path.
            nested_key (str): root field containing the serialized object.
        """
        with open(file_path, "w") as fp:
            yaml.dump({nested_key: self.to_dict()}, fp)


class ModelLoader(abc.ABC, Serializable):
    """Loads a machine learning model from somewhere."""

    def __init__(self, model_uri: str) -> None:
        super().__init__()
        self.model_uri = model_uri

    @abc.abstractmethod
    def __call__(self) -> MLModel:
        """Loads model from model URI."""



import os
import sys
from typing import Union, Dict, Optional, Any

import torch
from torch import Tensor
import lightning as pl
from lightning.pytorch.cli import LightningCLI

from itwinai.components import Trainer, Predictor, monitor_exec
from itwinai.serialization import ModelLoader
from itwinai.torch.inference import TorchModelLoader
from itwinai.torch.type import Batch
from itwinai.utils import load_yaml
from itwinai.torch.mlflow import (
    init_lightning_mlflow,
    teardown_lightning_mlflow
)


from model import ThreeDGAN
from dataloader import ParticlesDataModule


class Lightning3DGANTrainer(Trainer):
    def __init__(self, config: Union[Dict, str], exp_root: str = '.'):
        self.save_parameters(**self.locals2params(locals()))
        super().__init__()
        if isinstance(config, str) and os.path.isfile(config):
            # Load from YAML
            config = load_yaml(config)
        self.conf = config
        self.exp_root = exp_root

    @monitor_exec
    def execute(self) -> Any:
        init_lightning_mlflow(
            self.conf,
            tmp_dir=os.path.join(self.exp_root, '.tmp'),
            registered_model_name='3dgan-lite'
        )
        old_argv = sys.argv
        sys.argv = ['some_script_placeholder.py']
        cli = LightningCLI(
            args=self.conf,
            model_class=ThreeDGAN,
            datamodule_class=ParticlesDataModule,
            run=False,
            save_config_kwargs={
                "overwrite": True,
                "config_filename": "pl-training.yml",
            },
            subclass_mode_model=True,
            subclass_mode_data=True,
        )
        sys.argv = old_argv
        cli.trainer.fit(cli.model, datamodule=cli.datamodule)
        teardown_lightning_mlflow()


class LightningModelLoader(TorchModelLoader):
    """Loads a torch lightning model from somewhere.

    Args:
        model_uri (str): Can be a path on local filesystem
            or an mlflow 'locator' in the form:
            'mlflow+MLFLOW_TRACKING_URI+RUN_ID+ARTIFACT_PATH'
    """

    def __call__(self) -> pl.LightningModule:
        """"Loads model from model URI.

        Raises:
            ValueError: if the model URI is not recognized
                or the model is not found.

        Returns:
            pl.LightningModule: torch lightning module.
        """
        # TODO: improve
        # # Load best model
        # loaded_model = cli.model.load_from_checkpoint(
        #     ckpt_path,
        #     lightning_conf['model']['init_args']
        # )
        return super().__call__()


class Lightning3DGANPredictor(Predictor):

    def __init__(
        self,
        model: Union[ModelLoader, pl.LightningModule],
        config: Union[Dict, str],
        name: Optional[str] = None
    ):
        self.save_parameters(**self.locals2params(locals()))
        super().__init__(model, name)
        if isinstance(config, str) and os.path.isfile(config):
            # Load from YAML
            config = load_yaml(config)
        self.conf = config

    @monitor_exec
    def execute(
        self,
        datamodule: Optional[pl.LightningDataModule] = None,
        model: Optional[pl.LightningModule] = None
    ) -> Dict[str, Tensor]:
        old_argv = sys.argv
        sys.argv = ['some_script_placeholder.py']
        cli = LightningCLI(
            args=self.conf,
            model_class=ThreeDGAN,
            datamodule_class=ParticlesDataModule,
            run=False,
            save_config_kwargs={
                "overwrite": True,
                "config_filename": "pl-training.yml",
            },
            subclass_mode_model=True,
            subclass_mode_data=True,
        )
        sys.argv = old_argv

        # Override config file with inline arguments, if given
        if datamodule is None:
            datamodule = cli.datamodule
        if model is None:
            model = cli.model

        predictions = cli.trainer.predict(model, datamodule=datamodule)

        # Transpose predictions into images, energies and angles
        images = torch.cat(list(map(
            lambda pred: self.transform_predictions(
                pred['images']), predictions
        )))
        energies = torch.cat(list(map(
            lambda pred: pred['energies'], predictions
        )))
        angles = torch.cat(list(map(
            lambda pred: pred['angles'], predictions
        )))

        predictions_dict = dict()
        for img, en, ang in zip(images, energies, angles):
            sample_key = f"energy={en.item()}&angle={ang.item()}"
            predictions_dict[sample_key] = img

        return predictions_dict

    def transform_predictions(self, batch: Batch) -> Batch:
        """
        Post-process the predictions of the torch model.
        """
        return batch.squeeze(1)



"""
In the previous tutorial we saw how to create new components and assemble them
into a Pipeline for a simplified workflow execution. The Pipeline executes
the components in the order in which they are given, *assuming* that the
outputs of a component will fit as inputs of the following component.
This is not always true, thus you can use the ``Adapter`` component to
compensate for mismatches. This component allows to define a policy to
rearrange intermediate results between two components.

Moreover, it is good for reproducibility to keep track of the pipeline
configuration used to achieve some outstanding ML results. It would be a shame
to forget how you achieved state-of-the-art results!

itwinai allows to export the Pipeline form Python code to configuration file,
to persist both parameters and workflow structure. Exporting to configuration
file assumes that each component class resides in a separate python file, so
that the pipeline configuration is agnostic from the current python script.

Once the Pipeline has been exported to configuration file (YAML), it can
be executed directly from CLI:

>>> itwinai exec-pipeline --config my-pipeline.yaml --override nested.key=42

The itwinai CLI allows for dynamic override of configuration fields, by means
of nested key notation. Also list indices are supported:

>>> itwinai exec-pipeline --config my-pipe.yaml --override nested.list.2.0=42

"""
import subprocess
from itwinai.pipeline import Pipeline
from itwinai.parser import ConfigParser
from itwinai.components import Adapter

from basic_components import (
    MyDataGetter, MyDatasetSplitter, MyTrainer, MySaver
)

if __name__ == "__main__":

    # In this pipeline, the MyTrainer produces 4 elements as output: train,
    # validation, test datasets, and trained model. The Adapter selects the
    # trained model only, and forwards it to the saver, which expects a single
    # item as input.
    pipeline = Pipeline([
        MyDataGetter(data_size=100),
        MyDatasetSplitter(
            train_proportion=.5,
            validation_proportion=.25,
            test_proportion=0.25
        ),
        MyTrainer(),
        Adapter(policy=[f"{Adapter.INPUT_PREFIX}-1"]),
        MySaver()
    ])

    # Run pipeline
    trained_model = pipeline.execute()
    print("Trained model: ", trained_model)
    print("\n" + "="*50 + "\n")

    # Serialize pipeline to YAML
    pipeline.to_yaml("basic_pipeline_example.yaml", "pipeline")

    # Below, we show how to run a pre-existing pipeline stored as
    # a configuration file, with the possibility of dynamically
    # override some fields

    # Load pipeline from saved YAML (dynamic serialization)
    parser = ConfigParser(
        config="basic_pipeline_example.yaml",
        override_keys={
            "pipeline.init_args.steps.0.init_args.data_size": 200
        }
    )
    pipeline = parser.parse_pipeline()
    print(f"MyDataGetter's data_size is now: {pipeline.steps[0].data_size}\n")

    # Run parsed pipeline, with new data_size for MyDataGetter
    trained_model = pipeline.execute()
    print("Trained model (2): ", trained_model)

    # Save new pipeline to YAML file
    pipeline.to_yaml("basic_pipeline_example_v2.yaml", "pipeline")

    print("\n" + "="*50 + "\n")

    # Emulate pipeline execution from CLI, with dynamic override of
    # pipeline configuration fields
    subprocess.run(
        ["itwinai", "exec-pipeline", "--config",
         "basic_pipeline_example_v2.yaml",
         "--override",
         "pipeline.init_args.steps.0.init_args.data_size=300",
         "--override",
         "pipeline.init_args.steps.1.init_args.train_proportion=0.4"
         ]
    )



from typing import Optional
from ..components import BaseComponent, monitor_exec


class FakeGetter(BaseComponent):
    def __init__(self, data_uri: str, name: Optional[str] = None
                 ) -> None:
        super().__init__(name)
        self.save_parameters(data_uri=data_uri, name=name)
        self.data_uri = data_uri

    def execute(self):
        ...


class FakeGetterExec(FakeGetter):
    result: str = "dataset"

    @monitor_exec
    def execute(self):
        return self.result


class FakeSplitter(BaseComponent):
    def __init__(self, train_prop: float, name: Optional[str] = None
                 ) -> None:
        super().__init__(name)
        self.save_parameters(train_prop=train_prop, name=name)
        self.train_prop = train_prop

    def execute(self):
        ...


class FakeSplitterExec(FakeSplitter):
    result: tuple = ("train_dataset", "val_dataset", "test_dataset")

    @monitor_exec
    def execute(self, dataset):
        return self.result


class FakePreproc(BaseComponent):
    def __init__(self, max_items: int, name: Optional[str] = None
                 ) -> None:
        super().__init__(name)
        self.save_parameters(max_items=max_items, name=name)
        self.max_items = max_items

    def execute(self):
        ...


class FakePreprocExec(FakePreproc):
    @monitor_exec
    def execute(self, train_dataset, val_dataset, test_dataset):
        return train_dataset, val_dataset, test_dataset


class FakeTrainer(BaseComponent):
    def __init__(
        self,
        lr: float,
        batch_size: int,
        name: Optional[str] = None
    ) -> None:
        super().__init__(name)
        self.save_parameters(lr=lr, batch_size=batch_size, name=name)
        self.lr = lr
        self.batch_size = batch_size

    def execute(self):
        ...


class FakeTrainerExec(FakeTrainer):
    model: str = "trained_model"

    @monitor_exec
    def execute(self, train_dataset, val_dataset, test_dataset):
        return train_dataset, val_dataset, test_dataset, self.model


class FakeSaver(BaseComponent):
    def __init__(self, save_path: str, name: Optional[str] = None) -> None:
        super().__init__(name)
        self.save_parameters(save_path=save_path, name=name)
        self.save_path = save_path

    def execute(self):
        ...


class FakeSaverExec(FakeSaver):
    @monitor_exec
    def execute(self, artifact):
        return artifact



"""
Scaling test of Horovod on Imagenet using Resnet.
"""
from typing import Optional
import argparse
import os
import sys
from timeit import default_timer as timer
import time

import torch
# import torch.multiprocessing as mp
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler
import horovod.torch as hvd
import torchvision

from itwinai.parser import ArgumentParser as ItAIArgumentParser
from itwinai.loggers import EpochTimeTracker
from itwinai.torch.reproducibility import (
    seed_worker, set_seed
)

from utils import imagenet_dataset


def parse_params():
    parser = ItAIArgumentParser(description='PyTorch Imagenet Example')

    # Data and logging
    parser.add_argument('--data-dir', default='./',
                        help=('location of the training dataset in the '
                              'local filesystem'))
    parser.add_argument('--log-int', type=int, default=100,
                        help=('#batches to wait before logging training '
                              'status. Disabled if < 0.'))
    parser.add_argument('--verbose',
                        action=argparse.BooleanOptionalAction,
                        help='Print parsed arguments')
    parser.add_argument('--nworker', type=int, default=0,
                        help=('number of workers in DataLoader '
                              '(default: 0 - only main)'))
    parser.add_argument('--prefetch', type=int, default=2,
                        help='prefetch data in DataLoader (default: 2)')

    # Model
    parser.add_argument('--batch-size', type=int, default=64,
                        help='input batch size for training (default: 64)')
    parser.add_argument('--epochs', type=int, default=10,
                        help='number of epochs to train (default: 10)')
    parser.add_argument('--lr', type=float, default=0.01,
                        help='learning rate (default: 0.01)')
    parser.add_argument('--momentum', type=float, default=0.5,
                        help='SGD momentum (default: 0.5)')
    parser.add_argument('--shuff', action='store_true', default=False,
                        help='shuffle dataset (default: False)')

    # Reproducibility
    parser.add_argument('--rnd-seed', type=Optional[int], default=None,
                        help='seed integer for reproducibility (default: 0)')

    # Distributed ML
    parser.add_argument('--no-cuda', action='store_true', default=False,
                        help='disables CUDA training')
    parser.add_argument('--fp16-allreduce', action='store_true', default=False,
                        help='use fp16 compression during allreduce')
    parser.add_argument('--use-adasum', action='store_true', default=False,
                        help='use adasum algorithm to do reduction')
    parser.add_argument('--gradient-predivide-factor', type=float, default=1.0,
                        help=('apply gradient pre-divide factor in optimizer '
                              '(default: 1.0)'))

    args = parser.parse_args()
    if args.verbose:
        args_list = [f"{key}: {val}" for key, val in args.items()]
        print("PARSED ARGS:\n", '\n'.join(args_list))

    return args


def train(
        model, optimizer, train_sampler, train_loader,
        args, use_cuda, epoch, grank
):
    model.train()
    t_list = []
    loss_acc = 0
    if grank == 0:
        print("\n")
    for batch_idx, (data, target) in enumerate(train_loader):
        # if hvd.local_rank() == 0 and hvd.rank() == 0:
        #     print(f"BS == DATA: {data.shape}, TARGET: {target.shape}")
        t = timer()
        if use_cuda:
            data, target = data.cuda(), target.cuda()
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
        if grank == 0 and args.log_int > 0 and batch_idx % args.log_int == 0:
            # Use train_sampler to determine the number of examples in
            # this worker's partition
            print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                epoch, batch_idx * len(data), len(train_sampler),
                100. * batch_idx / len(train_loader), loss.item()))
        t_list.append(timer() - t)
        loss_acc += loss.item()
    if grank == 0:
        print('TIMER: train time', sum(t_list) / len(t_list), 's')
    return loss_acc


def main():
    # Parse CLI args
    args = parse_params()

    # Check resources availability
    use_cuda = not args.no_cuda and torch.cuda.is_available()
    is_distributed = False
    if use_cuda and torch.cuda.device_count() > 0:
        is_distributed = True

    # Start the time.time for profiling
    st = timer()

    if is_distributed:
        # Initializes the distributed backend which will
        # take care of synchronizing the workers (nodes/GPUs)
        hvd.init()

    # Set random seed for reproducibility
    torch_prng = set_seed(args.rnd_seed, deterministic_cudnn=False)

    # is_main_worker = True
    # if is_distributed and (hvd.rank() != 0 or hvd.local_rank() != 0):
    #     is_main_worker = False

    # Get local rank
    if is_distributed:
        lrank = hvd.local_rank()
        grank = hvd.rank()
        gwsize = hvd.size()
        lwsize = torch.cuda.device_count()
    else:
        # Use a single worker (either on GPU or CPU)
        lrank = 0
        grank = 0
        gwsize = 1
        lwsize = 1

    if grank == 0:
        print('TIMER: initialise:', timer()-st, 's')
        print('DEBUG: local ranks:', lwsize, '/ global ranks:', gwsize)
        print('DEBUG: sys.version:', sys.version)
        print('DEBUG: args.data_dir:', args.data_dir)
        print('DEBUG: args.log_int:', args.log_int)
        print('DEBUG: args.nworker:', args.nworker)
        print('DEBUG: args.prefetch:', args.prefetch)
        print('DEBUG: args.batch_size:', args.batch_size)
        print('DEBUG: args.epochs:', args.epochs)
        print('DEBUG: args.lr:', args.lr)
        print('DEBUG: args.momentum:', args.momentum)
        print('DEBUG: args.shuff:', args.shuff)
        print('DEBUG: args.rnd_seed:', args.rnd_seed)
        print('DEBUG: args.no_cuda:', args.no_cuda)
        print('DEBUG: args.fp16_allreduce:', args.fp16_allreduce)
        print('DEBUG: args.use_adasum:', args.use_adasum)
        print('DEBUG: args.gradient_predivide_factor:',
              args.gradient_predivide_factor)
        if use_cuda:
            print('DEBUG: torch.cuda.is_available():',
                  torch.cuda.is_available())
            print('DEBUG: torch.cuda.current_device():',
                  torch.cuda.current_device())
            print('DEBUG: torch.cuda.device_count():',
                  torch.cuda.device_count())
            print('DEBUG: torch.cuda.get_device_properties(hvd.local_rank()):',
                  torch.cuda.get_device_properties(hvd.local_rank()))

    if use_cuda:
        # Pin GPU to local rank
        torch.cuda.set_device(lrank)

    # Limit # of CPU threads to be used per worker
    # torch.set_num_threads(1)

    # Dataset
    train_dataset = imagenet_dataset(args.data_dir)

    # kwargs = {}
    # # When supported, use 'forkserver' to spawn dataloader workers instead...
    # # issues with Infiniband implementations that are not fork-safe
    # if (args.nworker > 0 and hasattr(mp, '_supports_context')
    #     and
    #         mp._supports_context and
    #         'forkserver' in mp.get_all_start_methods()):
    #     kwargs['multiprocessing_context'] = 'forkserver'

    if is_distributed:
        # Use DistributedSampler to partition the training data
        # Since Horovod is not based on torch.distributed,
        # `num_replicas` and `rank` cannot be retrieved from the
        # current distributed group, thus they need to be provided explicitly.
        train_sampler = DistributedSampler(
            train_dataset, num_replicas=gwsize, rank=grank,
            shuffle=(args.shuff and args.rnd_seed is None)
        )
        train_loader = DataLoader(
            train_dataset, batch_size=args.batch_size,
            sampler=train_sampler, num_workers=args.nworker, pin_memory=True,
            persistent_workers=(args.nworker > 1),
            prefetch_factor=args.prefetch, generator=torch_prng,
            worker_init_fn=seed_worker
        )  # , **kwargs)
    else:
        train_loader = DataLoader(
            train_dataset, batch_size=args.batch_size, generator=torch_prng,
            worker_init_fn=seed_worker
        )  # , **kwargs)

    # Create CNN model
    model = torchvision.models.resnet152()

    if use_cuda:
        model.cuda()

    if is_distributed:
        # By default, Adasum doesn't need scaling up learning rate
        lr_scaler = hvd.size() if not args.use_adasum else 1
        # If using GPU Adasum allreduce, scale learning rate by local_size
        if args.use_adasum and hvd.nccl_built():
            lr_scaler = hvd.local_size()
        # Scale learning rate by lr_scaler
        args.lr *= lr_scaler

    optimizer = optim.SGD(model.parameters(), lr=args.lr,
                          momentum=args.momentum)

    if is_distributed:
        # Broadcast parameters & optimizer state
        hvd.broadcast_parameters(model.state_dict(), root_rank=0)
        hvd.broadcast_optimizer_state(optimizer, root_rank=0)

        # Compression algorithm
        compression = (
            hvd.Compression.fp16 if args.fp16_allreduce
            else hvd.Compression.none
        )

        # Wrap optimizer with DistributedOptimizer
        optimizer = hvd.DistributedOptimizer(
            optimizer,
            named_parameters=model.named_parameters(),
            compression=compression,
            op=hvd.Adasum if args.use_adasum else hvd.Average,
            gradient_predivide_factor=args.gradient_predivide_factor)

    if grank == 0:
        print('TIMER: broadcast:', timer()-st, 's')
        print('\nDEBUG: start training')
        print('--------------------------------------------------------')
        nnod = os.environ.get('SLURM_NNODES', 'unk')
        epoch_time_tracker = EpochTimeTracker(
            series_name="horovod-bl",
            csv_file=f"epochtime_horovod-bl_{nnod}N.csv"
        )

    et = timer()
    start_epoch = 1
    for epoch in range(start_epoch, args.epochs + 1):
        lt = timer()
        if is_distributed:
            # Inform the sampler that a new epoch started: shuffle
            # may be needed
            train_sampler.set_epoch(epoch)

        # Training
        train(model, optimizer, train_sampler,
              train_loader, args, use_cuda, epoch, grank)

        # Save first epoch timer
        if epoch == start_epoch:
            first_ep_t = timer()-lt

        # Final epoch
        if epoch + 1 == args.epochs:
            train_loader.last_epoch = True

        if grank == 0:
            print('TIMER: epoch time:', timer()-lt, 's')
            epoch_time_tracker.add_epoch_time(epoch-1, timer()-lt)

    if grank == 0:
        print('\n--------------------------------------------------------')
        print('DEBUG: training results:\n')
        print('TIMER: first epoch time:', first_ep_t, ' s')
        print('TIMER: last epoch time:', timer()-lt, 's')
        print('TIMER: average epoch time:', (timer()-et)/args.epochs, ' s')
        print('TIMER: total epoch time:', timer()-et, ' s')
        if epoch > 1:
            print('TIMER: total epoch-1 time:',
                  timer()-et-first_ep_t, ' s')
            print('TIMER: average epoch-1 time:',
                  (timer()-et-first_ep_t)/(args.epochs-1), ' s')
        if use_cuda:
            print('DEBUG: memory req:',
                  int(torch.cuda.memory_reserved(lrank)/1024/1024), 'MB')
            print('DEBUG: memory summary:\n\n',
                  torch.cuda.memory_summary(0))
        print(f'TIMER: final time: {timer()-st} s\n')

    time.sleep(1)
    print(f"<Hvd rank: {hvd.rank()}> - TRAINING FINISHED")


if __name__ == "__main__":
    main()
    sys.exit()



from typing import Optional
import lightning as L

from torchvision.datasets import MNIST
from torch.utils.data import DataLoader, random_split
from torchvision import transforms

from itwinai.components import DataGetter, monitor_exec


class LightningMNISTDownloader(DataGetter):
    def __init__(
        self,
        data_path: str,
        name: Optional[str] = None
    ) -> None:
        super().__init__(name)
        self.save_parameters(**self.locals2params(locals()))
        self.data_path = data_path
        self._downloader = MNISTDataModule(
            data_path=self.data_path, download=True,
            # Mock other args...
            batch_size=1, train_prop=.5,
        )

    @monitor_exec
    def execute(self) -> None:
        # Simulate dataset creation to force data download
        self._downloader.setup(stage='fit')
        self._downloader.setup(stage='test')
        self._downloader.setup(stage='predict')


class MNISTDataModule(L.LightningDataModule):
    def __init__(
        self,
        data_path: str,
        batch_size: int,
        train_prop: float,
        download: bool = True
    ) -> None:
        super().__init__()
        self.data_path = data_path
        self.download = download
        self.batch_size = batch_size
        self.train_prop = train_prop
        self.transform = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize((0.1307,), (0.3081,)),
            ]
        )

    def setup(self, stage=None):
        if stage == "fit":
            mnist_full = MNIST(
                self.data_path, train=True,
                download=self.download,
                transform=self.transform
            )
            n_train_samples = int(self.train_prop * len(mnist_full))
            n_val_samples = len(mnist_full) - n_train_samples
            self.mnist_train, self.mnist_val = random_split(
                mnist_full, [n_train_samples, n_val_samples]
            )

        if stage == "test":
            self.mnist_test = MNIST(
                self.data_path, train=False,
                download=self.download,
                transform=self.transform
            )

        if stage == "predict":
            self.mnist_predict = MNIST(
                self.data_path, train=False,
                download=self.download,
                transform=self.transform
            )

    def train_dataloader(self):
        return DataLoader(
            self.mnist_train, batch_size=self.batch_size, num_workers=4)

    def val_dataloader(self):
        return DataLoader(
            self.mnist_val, batch_size=self.batch_size, num_workers=4)

    def test_dataloader(self):
        return DataLoader(
            self.mnist_test, batch_size=self.batch_size, num_workers=4)

    def predict_dataloader(self):
        return DataLoader(
            self.mnist_predict, batch_size=self.batch_size, num_workers=4)



import pytest
import shutil
import numpy as np
from unittest.mock import patch, MagicMock

from itwinai.loggers import (
    ConsoleLogger,
    MLFlowLogger,
    WandBLogger,
    TensorBoardLogger,
    LoggersCollection
)


@pytest.fixture(scope="module")
def console_logger():
    yield ConsoleLogger(savedir='/tmp/console/test_mllogs', log_freq=1)
    shutil.rmtree('/tmp/console/test_mllogs', ignore_errors=True)


@pytest.fixture(scope="module")
def mlflow_logger():
    yield MLFlowLogger(
        savedir='/tmp/mlflow/test_mllogs',
        experiment_name='test_experiment',
        tracking_uri='file:///tmp/mlruns'
    )
    shutil.rmtree('/tmp/mlflow/test_mllogs', ignore_errors=True)
    shutil.rmtree('/tmp/mlruns', ignore_errors=True)


@pytest.fixture(scope="module")
def wandb_logger():
    yield WandBLogger(savedir='/tmp/wandb/test_mllogs',
                      project_name='test_project')
    shutil.rmtree('/tmp/wandb/test_mllogs', ignore_errors=True)


@pytest.fixture(scope="module")
def tensorboard_logger_tf():
    yield TensorBoardLogger(savedir='/tmp/tf_tb/test_mllogs',
                            framework='tensorflow')
    shutil.rmtree('/tmp/tf_tb/test_mllogs', ignore_errors=True)


@pytest.fixture(scope="module")
def tensorboard_logger_torch():
    yield TensorBoardLogger(savedir='/tmp/torch_tb/test_mllogs',
                            framework='pytorch')
    shutil.rmtree('/tmp/torch_tb/test_mllogs', ignore_errors=True)


@pytest.fixture
def loggers_collection(console_logger, mlflow_logger, wandb_logger):
    return LoggersCollection([console_logger, mlflow_logger, wandb_logger])


def test_console_logger_log(console_logger):
    console_logger.create_logger_context()
    with patch('builtins.print') as mocked_print:
        console_logger.log('test_value', 'test_identifier')
        mocked_print.assert_called_with(
            'ConsoleLogger: test_identifier = test_value')
    console_logger.destroy_logger_context()


def test_mlflow_logger_log(mlflow_logger):
    with patch('mlflow.log_metric') as mock_log_metric:
        mlflow_logger.create_logger_context()
        mlflow_logger.log(0.5, 'test_metric', kind='metric', step=1)
        mock_log_metric.assert_called_once_with(
            key='test_metric', value=0.5, step=1)
        mlflow_logger.destroy_logger_context()


def test_wandb_logger_log(wandb_logger):
    with patch('wandb.init') as mock_init, patch('wandb.log') as mock_log:
        mock_init.return_value = MagicMock()
        wandb_logger.create_logger_context()
        wandb_logger.log(0.5, 'test_metric', kind='metric')
        mock_log.assert_called_once_with(
            {'test_metric': 0.5}, commit=True)


def test_tensorboard_logger_log_tf(tensorboard_logger_tf):
    import tensorflow as tf

    with patch('tensorflow.summary.scalar') as mock_scalar, \
            patch('tensorflow.summary.image') as mock_image, \
            patch('tensorflow.summary.text') as mock_text:
        tensorboard_logger_tf.create_logger_context()

        # Log a scalar
        tensorboard_logger_tf.log(0.5, 'test_scalar', kind='metric', step=1)
        mock_scalar.assert_called_once_with('test_scalar', 0.5, step=1)

        # Log an image
        image = tf.zeros([10, 10, 3])
        tensorboard_logger_tf.log(image, 'test_image', kind='image', step=1)
        mock_image.assert_called_once()
        assert np.allclose(mock_image.call_args[0][1].numpy(), image.numpy())

        # Log text
        tensorboard_logger_tf.log(
            'test text', 'test_text', kind='text', step=1)
        mock_text.assert_called_once_with('test_text', 'test text', step=1)

        tensorboard_logger_tf.destroy_logger_context()


def test_tensorboard_logger_log_torch(tensorboard_logger_torch):
    import torch

    step = 1
    with patch('torch.utils.tensorboard.SummaryWriter.add_scalar') \
            as mock_scalar, \
            patch('torch.utils.tensorboard.SummaryWriter.add_image') \
            as mock_image, \
            patch('torch.utils.tensorboard.SummaryWriter.add_text') \
            as mock_text:
        tensorboard_logger_torch.create_logger_context()

        # Log a scalar
        tensorboard_logger_torch.log(
            0.5, 'test_scalar', kind='metric', step=step)
        mock_scalar.assert_called_once_with(
            'test_scalar', 0.5, global_step=step)

        # Log an image
        image = torch.zeros([3, 10, 10])
        tensorboard_logger_torch.log(
            image, 'test_image', kind='image', step=step)
        mock_image.assert_called_once_with(
            'test_image', image, global_step=step)

        # Log text
        tensorboard_logger_torch.log(
            'test text', 'test_text', kind='text', step=step)
        mock_text.assert_called_once_with(
            'test_text', 'test text', global_step=step)

        tensorboard_logger_torch.destroy_logger_context()


def test_loggers_collection_log(loggers_collection):
    with patch('builtins.print') as mocked_print, \
            patch('mlflow.log_metric') as mock_log_metric, \
            patch('wandb.init') as mock_wandb_init, \
            patch('wandb.log') as mock_wandb_log:  # , \
        # patch('shutil.copyfile') as mock_shutil_copyfile, \
        # patch('pickle.dump') as mock_pickle_dump, \
        # patch('torch.save') as mock_torch_save:

        mock_wandb_init.return_value = MagicMock()

        loggers_collection.create_logger_context()
        loggers_collection.log(0.5, 'test_metric', kind='metric', step=1)

        mocked_print.assert_called_with('ConsoleLogger: test_metric = 0.5')
        mock_log_metric.assert_called_once_with(
            key='test_metric', value=0.5, step=1)
        mock_wandb_log.assert_called_once_with(
            {'test_metric': 0.5}, commit=True)

        loggers_collection.destroy_logger_context()



import os
import pytest


@pytest.fixture
def torch_env() -> str:
    """
    If TORCH_ENV env variable is defined, it overrides the default
    torch virtual environment name. Otherwise, fall back
    to './.venv-pytorch'.

    Returns absolute path to torch virtual environment.
    """
    if os.environ.get('TORCH_ENV') is None:
        env_p = './.venv-pytorch'
    else:
        env_p = str(os.environ.get('TORCH_ENV'))
    return os.path.abspath(env_p)


@pytest.fixture
def tf_env() -> str:
    """
    If TF_ENV env variable is defined, it overrides the default
    torch virtual environment name. Otherwise, fall back
    to './.venv-tf'.

    Returns absolute path to torch virtual environment.
    """
    if os.environ.get('TF_ENV') is None:
        env_p = './.venv-tf'
    else:
        env_p = str(os.environ.get('TF_ENV'))
    return os.path.abspath(env_p)



"""
The most simple workflow that you can write is a sequential pipeline of steps,
where the outputs of a component are fed as input to the following component,
employing a scikit-learn-like Pipeline.

In itwinai, a step is also called "component" and is implemented by extending
the ``itwinai.components.BaseComponent`` class. Each component implements
the `execute(...)` method, which provides a unified interface to interact with
each component.

The aim of itwinai components is to provide reusable machine learning best
practices, and some common operations are already encoded in some abstract
components. Some examples are:
- ``DataGetter``: has no input and returns a dataset, collected from somewhere
(e.g., downloaded).
- ``DataSplitter``: splits an input dataset into train, validation and test.
- ``DataPreproc``: perform preprocessing on train, validation, and test
datasets.
- ``Trainer``: trains an ML model and returns the trained model.
- ``Saver``: saved an ML artifact (e.g., dataset, model) to disk.

In this tutorial you will see how to create new components and how they
are assembled into sequential pipelines. Newly created components are
in a separate file called 'basic_components.py'.
"""
from itwinai.pipeline import Pipeline

# Import the custom components from file
from basic_components import MyDataGetter, MyDatasetSplitter, MyTrainer

if __name__ == "__main__":
    # Assemble them in a scikit-learn like pipeline
    pipeline = Pipeline([
        MyDataGetter(data_size=100),
        MyDatasetSplitter(
            train_proportion=.5,
            validation_proportion=.25,
            test_proportion=0.25
        ),
        MyTrainer()
    ])

    # Inspect steps
    print(pipeline[0])
    print(pipeline[2].name)
    print(pipeline[1].train_proportion)

    # Run pipeline
    _, _, _, trained_model = pipeline.execute()
    print("Trained model: ", trained_model)

    # You can also create a Pipeline from a dict of components, which
    # simplifies their retrieval by name
    pipeline = Pipeline({
        "datagetter": MyDataGetter(data_size=100),
        "splitter": MyDatasetSplitter(
            train_proportion=.5,
            validation_proportion=.25,
            test_proportion=0.25
        ),
        "trainer": MyTrainer()
    })

    # Inspect steps
    print(pipeline["datagetter"])
    print(pipeline["trainer"].name)
    print(pipeline["splitter"].train_proportion)

    # Run pipeline
    _, _, _, trained_model = pipeline.execute()
    print("Trained model: ", trained_model)



"""
This module provides the base classes to define modular and reproducible ML
workflows. The base component classes provide a template to follow for
extending existing components or creating new ones.

There are two ways of creating workflows: simple and advanced workflows.

Simple workflows can be obtained by creating a sequence of components
wrapped in a Pipeline object, which executes them in cascade, passing the
output of a component as the input of the following one. It is responsibility
of the user to prevent mismatches among outputs and inputs of component
sequences. This pipeline can be configured
both in terms of parameters and structure, with a configuration file
representing the whole pipeline. This configuration file can be executed
using itwinai CLI without the need of python files.

Example:

>>> from itwinai.components import DataGetter, Saver
>>> from itwinai.pipeline import Pipeline
>>>
>>> my_pipe = Pipeline({"getter": DataGetter(...), "data_saver": Saver(...)})
>>> my_pipe.execute()
>>> my_pipe.to_yaml("training_pipe.yaml")
>>>
>>> # The pipeline can be parsed back to Python with:
>>> from itwinai.parser import PipeParser
>>> my_pipe = PipeParser("training_pipe.yaml")
>>> my_pipe.execute()
>>>
>>> # Run the pipeline from configuration file with dynamic override
>>> itwinai exec-pipeline --config training_pipe.yaml \
>>> --override pipeline.init_args.steps.data_saver.some_param 42


Advanced workflows foresee more complicated connections between the
components and it is very difficult to define a structure beforehand
without risking of over-constraining the user. Therefore, advanced
workflows are defined by explicitly connecting component outputs to
to the inputs of other components, without a wrapper Pipeline object.
In this case, the configuration files enable the user to persist the
parameters passed to the argument parser, enabling reuse through
configuration files, with the possibility of dynamic overrides of parameters.

Example:

>>> from jsonargparse import ArgumentParser, ActionConfigFile
>>>
>>> parser = ArgumentParser(description='PyTorch MNIST Example')
>>> parser.add_argument('--batch-size', type=int, default=64,
>>>                     help='input batch size for training (default: 64)')
>>> parser.add_argument('--epochs', type=int, default=10,
>>>                     help='number of epochs to train (default: 10)')
>>> parser.add_argument('--lr', type=float, default=0.01,
>>>                     help='learning rate (default: 0.01)')
>>> parser.add_argument(
>>>     "-c", "--config", action=ActionConfigFile,
>>>     required=True,
>>>     help="Path to a configuration file in json or yaml format."
>>> )
>>> args = parser.parse_args()
>>>
>>> from itwinai.components import (
>>>     DataGetter, Saver, DataSplitter, Trainer
>>> )
>>> getter = DataGetter(...)
>>> splitter = DataSplitter(...)
>>> data_saver = Saver(...)
>>> model_saver = Saver(...)
>>> trainer = Trainer(
>>>     batch_size=args.batch_size, lr=args.lr, epochs=args.epochs
>>> )
>>>
>>> # Compose workflow
>>> my_dataset = getter.execute()
>>> train_set, valid_set, test_set = splitter.execute(my_dataset)
>>> data_saver.execute("train_dataset.pkl", test_set)
>>> _, _, _, trained_model = trainer(train_set, valid_set)
>>> model_saver.execute(trained_model)
>>>
>>> # Run the script using a previous configuration with dynamic override
>>> python my_train.py --config training_pipe.yaml --lr 0.002
"""


from __future__ import annotations
from typing import Any, Optional, Tuple, Union, Callable, Dict, List
from abc import ABC, abstractmethod
import time
import functools
# import logging
# from logging import Logger as PythonLogger

from .type import MLModel, MLDataset, MLArtifact
from .serialization import ModelLoader, Serializable
from .distributed import (
    detect_distributed_environment,
    distributed_patch_print
)


def monitor_exec(method: Callable) -> Callable:
    """Decorator for execute method of a component class.
    Computes execution time and gives some information about
    the execution of the component.

    Args:
        func (Callable): class method.
    """
    @functools.wraps(method)
    def monitored_method(self: BaseComponent, *args, **kwargs) -> Any:
        # Disable print in workers different from the main one,
        # when in distributed environments.
        dist_grank = detect_distributed_environment().global_rank
        distributed_patch_print(is_main=dist_grank == 0)

        msg = f"Starting execution of '{self.name}'..."
        self._printout(msg)
        start_t = time.time()
        try:
            # print(f'ARGS: {args}')
            # print(f'KWARGS: {kwargs}')
            result = method(self, *args, **kwargs)
        finally:
            self.cleanup()
        self.exec_t = time.time() - start_t
        msg = f"'{self.name}' executed in {self.exec_t:.3f}s"
        self._printout(msg)
        return result

    return monitored_method


class BaseComponent(ABC, Serializable):
    """Base component class. Each component provides a simple interface
    to foster modularity in machine learning code. Each component class
    implements the `execute` method, which received some input ML artifacts
    (e.g., datasets), performs some operations and returns new artifacts.
    The components are meant to be assembled in complex ML workflows,
    represented as pipelines.

        Args:
            name (Optional[str], optional): unique identifier for a step.
                Defaults to None.
        """
    _name: str = None
    #: Dictionary storing constructor arguments. Needed to serialize the
    #: class to dictionary. Set by ``self.save_parameters()`` method.
    parameters: Dict[Any, Any] = None

    def __init__(
        self,
        name: Optional[str] = None,
        # logs_dir: Optional[str] = None,
        # debug: bool = False,
    ) -> None:
        self.save_parameters(**self.locals2params(locals()))
        self.name = name

    @property
    def name(self) -> str:
        """Name of current component. Defaults to ``self.__class__.__name__``.
        """
        return (
            self._name if self._name is not None else self.__class__.__name__
        )

    @name.setter
    def name(self, name: str) -> None:
        self._name = name

    @abstractmethod
    @monitor_exec
    def execute(self, *args, **kwargs) -> Any:
        """"Execute some operations."""

    # def setup_console(self):
    #     """Setup Python logging"""
    #     self.log_file = os.path.join(self.logs_dir, self.name + ".log")
    #     f_handler = logging.FileHandler(self.log_file, mode='w')
    #     stdout_h = logging.StreamHandler(sys.stdout)

    #     if self.debug:
    #         log_format = ("%(asctime)s %(levelname)s "
    #                       "[%(filename)s:%(lineno)s - %(funcName)s()]: "
    #                       "%(message)s")
    #     else:
    #         log_format = ("%(levelname)s : %(message)s")

    #     logging.basicConfig(
    #         level=logging.DEBUG if self.debug else logging.INFO,
    #         handlers=[f_handler, stdout_h],
    #         format=log_format,
    #         datefmt="%Y-%m-%d %H:%M:%S"
    #     )
    #     self.console = logging.getLogger(self.name)

    def cleanup(self):
        """Cleanup resources allocated by this component."""

    @staticmethod
    def _printout(msg: str):
        msg = f"# {msg} #"
        print("#"*len(msg))
        print(msg)
        print("#"*len(msg))


class DataGetter(BaseComponent):
    """Retrieves a dataset."""

    @abstractmethod
    @monitor_exec
    def execute(self) -> MLDataset:
        """Retrieves a dataset.

        Returns:
            MLDataset: retrieved dataset.
        """


class DataProcessor(BaseComponent):
    """Performs dataset pre-processing."""

    @abstractmethod
    @monitor_exec
    def execute(
        self,
        train_dataset: MLDataset,
        validation_dataset: MLDataset,
        test_dataset: MLDataset
    ) -> Tuple[MLDataset, MLDataset, MLDataset]:
        """Trains a machine learning model.

        Args:
            train_dataset (MLDataset): training dataset.
            validation_dataset (MLDataset): validation dataset.
            test_dataset (MLDataset): test dataset.

        Returns:
            Tuple[MLDataset, MLDataset, MLDataset]: preprocessed training
            dataset, validation dataset, test dataset.
        """


class DataSplitter(BaseComponent):
    """Splits a dataset into train, validation, and test splits."""
    _train_proportion: Union[int, float]
    _validation_proportion: Union[int, float]
    _test_proportion: Union[int, float]

    def __init__(
        self,
        train_proportion: Union[int, float],
        validation_proportion: Union[int, float],
        test_proportion: Union[int, float],
        name: Optional[str] = None
    ) -> None:
        super().__init__(name)
        self.save_parameters(**self.locals2params(locals()))
        self.train_proportion = train_proportion
        self.validation_proportion = validation_proportion
        self.test_proportion = test_proportion

    @property
    def train_proportion(self) -> Union[int, float]:
        """Training set proportion."""
        return self._train_proportion

    @train_proportion.setter
    def train_proportion(self, prop: Union[int, float]) -> None:
        if isinstance(prop, float) and not 0.0 <= prop <= 1.0:
            raise ValueError(
                "Train proportion should be in the interval [0.0, 1.0] "
                f"if given as float. Received {prop}"
            )
        self._train_proportion = prop

    @property
    def validation_proportion(self) -> Union[int, float]:
        """Validation set proportion."""
        return self._validation_proportion

    @validation_proportion.setter
    def validation_proportion(self, prop: Union[int, float]) -> None:
        if isinstance(prop, float) and not 0.0 <= prop <= 1.0:
            raise ValueError(
                "Validation proportion should be in the interval [0.0, 1.0] "
                f"if given as float. Received {prop}"
            )
        self._validation_proportion = prop

    @property
    def test_proportion(self) -> Union[int, float]:
        """Test set proportion."""
        return self._test_proportion

    @test_proportion.setter
    def test_proportion(self, prop: Union[int, float]) -> None:
        if isinstance(prop, float) and not 0.0 <= prop <= 1.0:
            raise ValueError(
                "Test proportion should be in the interval [0.0, 1.0] "
                f"if given as float. Received {prop}"
            )
        self._test_proportion = prop

    @abstractmethod
    @monitor_exec
    def execute(
        self,
        dataset: MLDataset
    ) -> Tuple[MLDataset, MLDataset, MLDataset]:
        """Splits a dataset into train, validation and test splits.

        Args:
            dataset (MLDataset): input dataset.

        Returns:
            Tuple[MLDataset, MLDataset, MLDataset]: tuple of
            train, validation and test splits.
        """


class Trainer(BaseComponent):
    """Trains a machine learning model."""

    @abstractmethod
    @monitor_exec
    def execute(
        self,
        train_dataset: MLDataset,
        validation_dataset: MLDataset,
        test_dataset: MLDataset
    ) -> Tuple[MLDataset, MLDataset, MLDataset, MLModel]:
        """Trains a machine learning model.

        Args:
            train_dataset (MLDataset): training dataset.
            validation_dataset (MLDataset): validation dataset.
            test_dataset (MLDataset): test dataset.

        Returns:
            Tuple[MLDataset, MLDataset, MLDataset]: training dataset,
            validation dataset, test dataset, trained model.
        """


class Predictor(BaseComponent):
    """Applies a pre-trained machine learning model to unseen data."""

    #: Pre-trained ML model used to make predictions.
    model: MLModel

    def __init__(
        self,
        model: Union[MLModel, ModelLoader],
        name: Optional[str] = None,
    ) -> None:
        super().__init__(name=name)
        self.save_parameters(**self.locals2params(locals()))
        self.model = model() if isinstance(model, ModelLoader) else model

    @abstractmethod
    @monitor_exec
    def execute(
        self,
        predict_dataset: MLDataset,
        model: Optional[MLModel] = None
    ) -> MLDataset:
        """Applies a machine learning model on a dataset of samples.

        Args:
            predict_dataset (MLDataset): dataset for inference.
            model (Optional[MLModel], optional): overrides the internal model,
                if given. Defaults to None.

        Returns:
            MLDataset: predictions with the same cardinality of the
            input dataset.
        """


class Saver(BaseComponent):
    """Saves artifact to disk."""

    @abstractmethod
    @monitor_exec
    def execute(self, artifact: MLArtifact) -> MLArtifact:
        """Saves an ML artifact to disk.

        Args:
            artifact (MLArtifact): artifact to save.

        Returns:
            MLArtifact: the same input artifact, after saving it.
        """


class Adapter(BaseComponent):
    """Connects to components in a sequential pipeline, allowing to
    control with greater detail how intermediate results are propagated
    among the components.

    Args:
        policy (List[Any]): list of the same length of the output of this
            component, describing how to map the input args to the output.
        name (Optional[str], optional): name of the component.
            Defaults to None.

    The adapter allows to define a policy with which inputs are re-arranged
    before being propagated to the next component.
    Some examples: [policy]: (input) -> (output)

    - ["INPUT_ARG#2", "INPUT_ARG#1", "INPUT_ARG#0"]: (11,22,33) -> (33,22,11)

    - ["INPUT_ARG#0", "INPUT_ARG#2", None]: (11, 22, 33) -> (11, 33, None)

    - []: (11, 22, 33) -> ()

    - [42, "INPUT_ARG#2", "hello"] -> (11,22,33,44,55) -> (42, 33, "hello")

    - [None, 33, 3.14]: () -> (None, 33, 3.14)

    - [None, 33, 3.14]: ("double", 44, None, True) -> (None, 33, 3.14)

    """

    #: Adapter policy.
    policy: List[Any]
    INPUT_PREFIX: str = "INPUT_ARG#"

    def __init__(self, policy: List[Any], name: Optional[str] = None) -> None:
        super().__init__(name=name)
        self.save_parameters(**self.locals2params(locals()))
        self.name = name
        self.policy = policy

    @monitor_exec
    def execute(self, *args) -> Tuple:
        """Produces an output tuple by arranging input arguments according
        to the policy specified in the constructor.

        Args:
            args (Tuple): input arguments.

        Returns:
            Tuple: input args arranged according to some policy.
        """
        result = []
        for itm in self.policy:
            if isinstance(itm, str) and itm.startswith(self.INPUT_PREFIX):
                arg_idx = int(itm[len(self.INPUT_PREFIX):])
                if arg_idx >= len(args):
                    max_idx = max(map(
                        lambda itm: int(itm[len(self.INPUT_PREFIX):]),
                        filter(
                            lambda el: (
                                isinstance(el, str)
                                and el.startswith(self.INPUT_PREFIX)
                            ),
                            self.policy
                        )))
                    raise IndexError(
                        f"The args received as input by '{self.name}' "
                        "are not consistent with the given adapter policy "
                        "because input args are too few! "
                        f"Input args are {len(args)} but the policy foresees "
                        f"at least {max_idx+1} items."
                    )
                result.append(args[arg_idx])
            else:
                result.append(itm)
        return tuple(result)



import tensorflow as tf
import joblib
import time

from .macros import Network
from cyclones_vgg import (
    custom_VGG_V1, custom_VGG_V2, custom_VGG_V3  # , ModelV5
)


def saveparams(file, **kwargs):
    joblib.dump(kwargs, file)


def readparams(file):
    return joblib.load(file)


class Timer():

    def __init__(
        self,
        timers=['tot_exec_elapsed_time',
                'io_elapsed_time', 'training_elapsed_time']
    ):
        # initialize execution times data structure
        self.exec_times = {}
        self.partials = {}
        for t in timers:
            self.exec_times.update({t: 0})
            self.partials.update({t: 0})

    def start(self, timer):
        # update partial timers to start counting
        self.partials.update({timer: -time.time()})

    def stop(self, timer):
        # add ( stop - start ) time to global execution time
        self.exec_times[timer] += self.partials[timer] + time.time()
        # reset partial
        self.partials[timer] = 0


def get_network_config(network, **kwargs):
    # choose the network configuration based on the passed network type
    if network == Network.VGG_V1.value:
        print('Using custom VGG V1')
        model = custom_VGG_V1(
            patch_size=kwargs['patch_size'], channels=kwargs['channels'],
            activation=kwargs['activation'], regularizer=kwargs['regularizer'])

    elif network == Network.VGG_V2.value:
        print('Using custom VGG V2')
        model = custom_VGG_V2(
            patch_size=kwargs['patch_size'], channels=kwargs['channels'],
            activation=kwargs['activation'], regularizer=kwargs['regularizer'])

    elif network == Network.VGG_V3.value:
        print('Using custom VGG V3')
        model = custom_VGG_V3(
            patch_size=kwargs['patch_size'], channels=kwargs['channels'],
            activation=kwargs['activation'], regularizer=kwargs['regularizer'])

    # elif network == Network.MODEL_V5.value:
    #     print('Using Model V5')
    #     model = ModelV5(
    #         patch_size=kwargs['patch_size'], channels=kwargs['channels'],
    #         last_activation=kwargs['activation'],
    #         kernel_size=kwargs['kernel_size'])

    return model


def load_model(model_fpath):
    """
    Loads a keras model from a file, recognizing whether or not it is a weight
    file or a model file.
    """
    model_fname = model_fpath.split('/')[-1]
    if 'model' in model_fname:
        try:
            model = tf.keras.models.load_model(model_fpath)
        except Exception as e:
            print(f'Cannot load model. Caused by error: {e}')
    elif 'weight' in model_fname:
        model.compile()
        model.built = True
        model.load_weights(model_fpath)
    return model



import abc
import os
from pydantic import BaseModel
import builtins as __builtin__


class DistributedStrategy(abc.ABC):
    """Abstract class to define the distributed backend methods."""


class ClusterEnvironment(BaseModel):
    """Stores information about distributed environment."""
    #: Global rank of current worker, in a distributed environment.
    #: ``global_rank==0`` identifies the main worker.
    #: Defaults to 0.
    global_rank: int = 0
    #: Local rank of current worker, in a distributed environment.
    #: Defaults to 0.
    local_rank: int = 0
    #: Total number of workers in a distributed environment.
    #: Defaults to 1.
    global_world_size: int = 1
    #: Number of workers on the same node in a distributed environment.
    #: Defaults to 1.
    local_world_size: int = 1


def detect_distributed_environment() -> ClusterEnvironment:
    """Detects distributed environment, extracting information like
    global ans local ranks, and world size.
    """
    if os.getenv('TORCHELASTIC_RUN_ID') is not None:
        # Torch elastic environment
        # https://pytorch.org/docs/stable/elastic/run.html#environment-variables
        return ClusterEnvironment(
            global_rank=os.getenv('RANK'),
            local_rank=os.getenv('LOCAL_RANK'),
            local_world_size=os.getenv('LOCAL_WORLD_SIZE'),
            global_world_size=os.getenv('WORLD_SIZE')
        )
    elif os.getenv('OMPI_COMM_WORLD_SIZE') is not None:
        # Open MPI environment
        # https://docs.open-mpi.org/en/v5.0.x/tuning-apps/environment-var.html
        return ClusterEnvironment(
            global_rank=os.getenv('OMPI_COMM_WORLD_RANK'),
            local_rank=os.getenv('OMPI_COMM_WORLD_LOCAL_RANK'),
            local_world_size=os.getenv('OMPI_COMM_WORLD_LOCAL_SIZE'),
            global_world_size=os.getenv('OMPI_COMM_WORLD_SIZE')
        )
    elif os.getenv('SLURM_JOB_ID') is not None:
        print("WARNING: detected SLURM environment, but "
              "unable to determine ranks and world sizes!")
        return ClusterEnvironment()
    else:
        return ClusterEnvironment()


#: Save original builtin print before patching it in distributed environments
builtin_print = __builtin__.print


def distributed_patch_print(is_main: bool) -> None:
    """Disable ``print()`` when not in main worker.

    Args:
        is_main (bool): whether it is called from main worker.
    """

    def print(*args, **kwself):
        """Print is disables on workers different from
        the main one, unless the print is called with
        ``force=True`` argument.
        """
        force = kwself.pop('force', False)
        if is_main or force:
            builtin_print(*args, **kwself)

    # Patch builtin print
    __builtin__.print = print



import yaml
import pytest

from itwinai.components import BaseComponent
from itwinai.parser import ConfigParser, add_replace_field
from itwinai.tests import FakeTrainer, FakePreproc, FakeSaver


def test_add_replace_field():
    conf = {}
    add_replace_field(conf, "some.key.chain", 123)
    target1 = dict(some=dict(key=dict(chain=123)))
    assert conf == target1

    add_replace_field(conf, "some.key.chain", 222)
    target2 = dict(some=dict(key=dict(chain=222)))
    assert conf == target2

    add_replace_field(conf, "some.key.field", 333)
    target3 = dict(some=dict(key=dict(chain=222, field=333)))
    assert conf == target3

    conf['some']['list'] = [1, 2, 3]
    add_replace_field(conf, "some.list.0", 3)
    target4 = dict(some=dict(
        key=dict(chain=222, field=333),
        list=[3, 2, 3]
    ))
    assert conf == target4

    add_replace_field(conf, "some.list.0.some.el", 7)
    target5 = dict(some=dict(
        key=dict(chain=222, field=333),
        list=[dict(some=dict(el=7)), 2, 3]
    ))
    assert conf == target5

    conf2 = dict(first=dict(list1=[[0, 1], [2, 3]], el=0))
    add_replace_field(conf2, "first.list1.1.0", 77)
    target6 = dict(first=dict(list1=[[0, 1], [77, 3]], el=0))
    assert conf2 == target6

    conf3 = dict(first=dict(
        list1=[[0, dict(nst=("el", dict(ciao="ciao")))], [2, 3]], el=0))
    add_replace_field(conf3, "first.list1.0.1.nst.1.ciao", "hello")
    target7 = dict(first=dict(
        list1=[[0, dict(nst=("el", dict(ciao="hello")))], [2, 3]], el=0))
    assert conf3 == target7

    add_replace_field(conf3, "first.list1.0.1.nst.1.ciao.I.am.john", True)
    target8 = dict(first=dict(
        list1=[
            [0, dict(nst=("el", dict(ciao=dict(I=dict(am=dict(john=True))))))],
            [2, 3]
        ], el=0))
    assert conf3 == target8


def test_parse_list_pipeline():
    """Parse a pipeline from config file,
    where the pipeline is define as a list of components.
    """
    config = yaml.safe_load(pytest.PIPE_LIST_YAML)
    parser = ConfigParser(config=config)
    pipe = parser.parse_pipeline(
        pipeline_nested_key="my-list-pipeline"
    )

    assert isinstance(pipe.steps, list)
    for step in pipe.steps:
        assert isinstance(step, BaseComponent)


def test_parse_dict_pipeline():
    """Parse a pipeline from config file,
    where the pipeline is define as a dict of components.
    """
    config = yaml.safe_load(pytest.PIPE_DICT_YAML)
    parser = ConfigParser(config=config)
    pipe = parser.parse_pipeline(
        pipeline_nested_key="my-dict-pipeline"
    )

    assert isinstance(pipe.steps, dict)
    for step in pipe.steps.values():
        assert isinstance(step, BaseComponent)


def test_parse_non_existing_pipeline():
    """Parse a pipeline from config file,
    where the pipeline key is wrong.
    """
    config = yaml.safe_load(pytest.PIPE_DICT_YAML)
    parser = ConfigParser(config=config)
    with pytest.raises(KeyError):
        _ = parser.parse_pipeline(
            pipeline_nested_key="non-existing-pipeline"
        )


def test_parse_nested_pipeline():
    """Parse a pipeline from config file,
    where the pipeline key is nested.
    """
    config = yaml.safe_load(pytest.NESTED_PIPELINE)
    parser = ConfigParser(config=config)
    _ = parser.parse_pipeline(
        pipeline_nested_key="some.field.nst-pipeline"
    )


def test_dynamic_override_parser_pipeline_dict():
    """Parse a pipeline from config file,
    and verify that dynamic override works
    in a pipeline composed of a dict of components.
    """
    config = yaml.safe_load(pytest.PIPE_DICT_YAML)

    override_keys = {
        "my-dict-pipeline.init_args.steps.preproc-step.init_args.max_items": 33
    }
    parser = ConfigParser(config=config, override_keys=override_keys)
    pipe = parser.parse_pipeline(
        pipeline_nested_key="my-dict-pipeline"
    )
    assert pipe.steps['preproc-step'].max_items == 33


def test_dynamic_override_parser_pipeline_list():
    """Parse a pipeline from config file,
    and verify that dynamic override works
    in a pipeline composed of a list of components.
    """
    config = yaml.safe_load(pytest.PIPE_LIST_YAML)

    override_keys = {
        "my-list-pipeline.init_args.steps.0.init_args.max_items": 42
    }
    parser = ConfigParser(config=config, override_keys=override_keys)
    pipe = parser.parse_pipeline(
        pipeline_nested_key="my-list-pipeline"
    )
    assert pipe.steps[0].max_items == 42


def test_parse_step_list_pipeline():
    """Parse a pipeline step from config file,
    where the pipeline is define as a list of components.
    """
    config = yaml.safe_load(pytest.PIPE_LIST_YAML)
    parser = ConfigParser(config=config)
    step = parser.parse_step(
        step_idx=1,
        pipeline_nested_key="my-list-pipeline"
    )

    assert isinstance(step, BaseComponent)
    assert isinstance(step, FakeTrainer)

    with pytest.raises(IndexError):
        _ = parser.parse_step(
            step_idx=12,
            pipeline_nested_key="my-list-pipeline"
        )
    with pytest.raises(TypeError):
        _ = parser.parse_step(
            step_idx='my-step-name',
            pipeline_nested_key="my-list-pipeline"
        )


def test_parse_step_dict_pipeline():
    """Parse a pipeline step from config file,
    where the pipeline is define as a dict of components.
    """
    config = yaml.safe_load(pytest.PIPE_DICT_YAML)
    parser = ConfigParser(config=config)
    step = parser.parse_step(
        step_idx='preproc-step',
        pipeline_nested_key="my-dict-pipeline"
    )

    assert isinstance(step, BaseComponent)
    assert isinstance(step, FakePreproc)

    with pytest.raises(KeyError):
        _ = parser.parse_step(
            step_idx='unk-step',
            pipeline_nested_key="my-dict-pipeline"
        )
    with pytest.raises(KeyError):
        _ = parser.parse_step(
            step_idx=0,
            pipeline_nested_key="my-dict-pipeline"
        )


def test_parse_step_nested_pipeline():
    """Parse a pipeline step from config file,
    where the pipeline is nested under some field.
    """
    config = yaml.safe_load(pytest.NESTED_PIPELINE)
    parser = ConfigParser(config=config)
    step = parser.parse_step(
        step_idx=2,
        pipeline_nested_key="some.field.nst-pipeline"
    )

    assert isinstance(step, BaseComponent)
    assert isinstance(step, FakeSaver)

    with pytest.raises(KeyError):
        _ = parser.parse_step(
            step_idx=2,
            pipeline_nested_key="my-pipeline"
        )



"""
Tests for MNIST use case.

Intended to be integration tests, to make sure that updates in the code base
do not break use cases' workflows.
"""

import pytest
import subprocess
import os
# from itwinai.cli import exec_pipeline

TORCH_PATH = "use-cases/mnist/torch"
LIGHTNING_PATH = "use-cases/mnist/torch-lightning"
TF_PATH = "use-cases/mnist/tensorflow"
DEFAULT_MNIST_DATASET = '.tmp'


@pytest.mark.skip(reason="structure changed")
def test_structure_mnist_torch(check_folder_structure):
    """Test MNIST folder structure for torch native trainer."""
    check_folder_structure(TORCH_PATH)


@pytest.mark.skip(reason="structure changed")
def test_structure_mnist_lightning(check_folder_structure):
    """Test MNIST folder structure for torch lightning trainer."""
    check_folder_structure(LIGHTNING_PATH)


@pytest.mark.skip(reason="structure changed")
def test_structure_mnist_tf(check_folder_structure):
    """Test MNIST folder structure for tensorflow trainer."""
    check_folder_structure(TF_PATH)


@pytest.mark.functional
def test_mnist_train_torch(torch_env, tmp_test_dir, install_requirements):
    """
    Test MNIST torch native trainer by running it end-to-end.

    If MNIST_DATASET env variable is defined, it is used to
    override the default dataset download location: useful
    when it contains a local copy of the dataset, preventing
    downloading it again.
    """
    install_requirements(TORCH_PATH, torch_env)

    if os.environ.get('MNIST_DATASET'):
        dataset_path = os.environ.get('MNIST_DATASET')
    else:
        dataset_path = DEFAULT_MNIST_DATASET
    conf = os.path.join(os.path.abspath(TORCH_PATH), 'config.yaml')
    cmd = (f"{torch_env}/bin/itwinai exec-pipeline "
           f"--config {conf} --pipe-key training_pipeline "
           f"-o dataset_root={dataset_path}")
    subprocess.run(cmd.split(), check=True, cwd=tmp_test_dir)


@pytest.mark.functional
def test_mnist_inference_torch(torch_env, tmp_test_dir, install_requirements):
    """
    Test MNIST torch native inference by running it end-to-end.
    """
    install_requirements(TORCH_PATH, torch_env)

    # Create fake inference dataset and checkpoint
    exec = os.path.join(os.path.abspath(TORCH_PATH),
                        'create_inference_sample.py')
    cmd = (f"{torch_env}/bin/python {exec} "
           f"--root {tmp_test_dir}")
    subprocess.run(cmd.split(), check=True, cwd=tmp_test_dir)

    # Test inference
    conf = os.path.join(os.path.abspath(TORCH_PATH), 'config.yaml')
    cmd = (f"{torch_env}/bin/itwinai exec-pipeline "
           f"--config {conf} --pipe-key inference_pipeline")
    subprocess.run(cmd.split(), check=True, cwd=tmp_test_dir)


@pytest.mark.functional
def test_mnist_train_torch_lightning(
    torch_env,
    tmp_test_dir,
    install_requirements
):
    """
    Test MNIST torch lightning trainer by running it end-to-end.

    If MNIST_DATASET env variable is defined, it is used to
    override the default dataset download location: useful
    when it contains a local copy of the dataset, preventing
    downloading it again.
    """
    install_requirements(LIGHTNING_PATH, torch_env)

    if os.environ.get('MNIST_DATASET'):
        dataset_path = os.environ.get('MNIST_DATASET')
    else:
        dataset_path = DEFAULT_MNIST_DATASET
    conf = os.path.join(os.path.abspath(LIGHTNING_PATH), 'config.yaml')
    cmd = (f"{torch_env}/bin/itwinai exec-pipeline "
           f"--config {conf} --pipe-key training_pipeline "
           f"-o dataset_root={dataset_path}")
    subprocess.run(cmd.split(), check=True, cwd=tmp_test_dir)


@pytest.mark.functional
def test_mnist_train_tf(tf_env, tmp_test_dir, install_requirements):
    """
    Test MNIST tensorflow trainer by running it end-to-end.
    """
    install_requirements(TF_PATH, tf_env)
    conf = os.path.join(os.path.abspath(TF_PATH), 'pipeline.yaml')
    cmd = (f"{tf_env}/bin/itwinai exec-pipeline "
           f"--config {conf} --pipe-key pipeline")
    subprocess.run(cmd.split(), check=True, cwd=tmp_test_dir)



"""
This module is used during inference to save predicted labels to file.
"""

from typing import Optional, List, Dict
import os
import shutil
import csv

from itwinai.components import Saver, monitor_exec


class TorchMNISTLabelSaver(Saver):
    """Serializes to disk the labels predicted for MNIST dataset."""

    def __init__(
        self,
        save_dir: str = 'mnist_predictions',
        predictions_file: str = 'predictions.csv',
        class_labels: Optional[List] = None
    ) -> None:
        super().__init__()
        self.save_parameters(**self.locals2params(locals()))
        self.save_dir = save_dir
        self.predictions_file = predictions_file
        self.class_labels = (
            class_labels if class_labels is not None
            else [f'Digit {i}' for i in range(10)]
        )

    @monitor_exec
    def execute(self, predicted_classes: Dict[str, int],) -> Dict[str, int]:
        """Translate predictions from class idx to class label and save
        them to disk.

        Args:
            predicted_classes (Dict[str, int]): maps unique item ID to
                the predicted class ID.

        Returns:
            Dict[str, int]: predicted classes.
        """
        if os.path.exists(self.save_dir):
            shutil.rmtree(self.save_dir)
        os.makedirs(self.save_dir)

        # Map class idx (int) to class label (str)
        predicted_labels = {
            itm_name: self.class_labels[cls_idx]
            for itm_name, cls_idx in predicted_classes.items()
        }

        # Save to disk
        filepath = os.path.join(self.save_dir, self.predictions_file)
        with open(filepath, 'w') as csv_file:
            writer = csv.writer(csv_file)
            for key, value in predicted_labels.items():
                writer.writerow([key, value])
        return predicted_labels



from .dummy_components import (
    FakeGetter, FakeGetterExec, FakePreproc, FakePreprocExec,
    FakeSaver, FakeSaverExec, FakeSplitter, FakeSplitterExec,
    FakeTrainer, FakeTrainerExec
)

_ = (
    FakeGetter, FakeGetterExec, FakePreproc, FakePreprocExec,
    FakeSaver, FakeSaverExec, FakeSplitter, FakeSplitterExec,
    FakeTrainer, FakeTrainerExec
)



"""Create a simple inference dataset sample and a checkpoint."""

import argparse
import os
import torch
from model import ThreeDGAN


def create_checkpoint(
    root: str = '.',
    ckpt_name: str = "3dgan-inference.pth"
):
    ckpt_path = os.path.join(root, ckpt_name)
    net = ThreeDGAN()
    torch.save(net, ckpt_path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--root", type=str, default='.')
    parser.add_argument("--ckpt-name", type=str, default="3dgan-inference.pth")
    args = parser.parse_args()
    create_checkpoint(**vars(args))



from typing import Dict, Optional
import os

import mlflow
import yaml


def _get_mlflow_logger_conf(pl_config: Dict) -> Optional[Dict]:
    """Extract MLFLowLogger configuration from pytorch lightning
    configuration file, if present.

    Args:
        pl_config (Dict): lightning configuration loaded in memory.

    Returns:
        Optional[Dict]: if present, MLFLowLogger constructor arguments
        (under 'init_args' key).
    """
    if not pl_config['trainer'].get('logger'):
        return None
    if isinstance(pl_config['trainer']['logger'], list):
        # If multiple loggers are provided
        for logger_conf in pl_config['trainer']['logger']:
            if logger_conf['class_path'].endswith('MLFlowLogger'):
                return logger_conf['init_args']
    elif pl_config['trainer']['logger']['class_path'].endswith('MLFlowLogger'):
        return pl_config['trainer']['logger']['init_args']


def _mlflow_log_pl_config(pl_config: Dict, local_yaml_path: str) -> None:
    os.makedirs(os.path.dirname(local_yaml_path), exist_ok=True)
    with open(local_yaml_path, 'w') as outfile:
        yaml.dump(pl_config, outfile, default_flow_style=False)
    mlflow.log_artifact(local_yaml_path)


def init_lightning_mlflow(
    pl_config: Dict,
    default_experiment_name: str = 'Default',
    tmp_dir: str = '.tmp',
    **autolog_kwargs
) -> None:
    """Initialize mlflow for pytorch lightning, also setting up
    auto-logging (mlflow.pytorch.autolog(...)). Creates a new mlflow
    run and attaches it to the mlflow auto-logger.

    Args:
        pl_config (Dict): pytorch lightning configuration loaded in memory.
        default_experiment_name (str, optional): used as experiment name
            if it is not given in the lightning conf. Defaults to 'Default'.
        tmp_dir (str): where to temporarily store some artifacts.
        autolog_kwargs (kwargs): args for mlflow.pytorch.autolog(...).
    """
    mlflow_conf: Optional[Dict] = _get_mlflow_logger_conf(pl_config)
    if not mlflow_conf:
        return

    tracking_uri = mlflow_conf.get('tracking_uri')
    if not tracking_uri:
        save_path = mlflow_conf.get('save_dir')
        tracking_uri = "file://" + os.path.abspath(save_path)

    experiment_name = mlflow_conf.get('experiment_name')
    if not experiment_name:
        experiment_name = default_experiment_name

    mlflow.set_tracking_uri(tracking_uri)
    mlflow.set_experiment(experiment_name)
    mlflow.pytorch.autolog(**autolog_kwargs)
    run = mlflow.start_run()
    print(f"MLFlow's artifacts URI: {run.info.artifact_uri}")

    mlflow_conf['experiment_name'] = experiment_name
    mlflow_conf['run_id'] = mlflow.active_run().info.run_id

    _mlflow_log_pl_config(pl_config, os.path.join(tmp_dir, 'pl_config.yml'))


def teardown_lightning_mlflow() -> None:
    """End active mlflow run, if any."""
    if mlflow.active_run() is not None:
        mlflow.end_run()



import psutil
import os
import time
import pandas as pd
import tensorflow as tf


class ProcessBenchmark(tf.keras.callbacks.Callback):
    """
    Gets several benchmarks parameters about the process during the training,
    including execution time.

    Parameters
    ----------
    filename : pathlike
        The file in which will be put the resulting history.

    """

    def __init__(self, filename):
        # get process pid
        self.pid = os.getpid()
        # instantiate a process object from which we'll get status information
        self.process = psutil.Process(self.pid)
        # get initial process cpu percent, so that the next call returns
        # the correct cpu percentage
        self.process.cpu_percent()

        self.filename = filename
        # create time history dataframe
        self.benchmark_df = pd.DataFrame(columns=[
            'time',
            'start_cpu_percent',
            'end_cpu_percent',
            'start_mem_rss',
            'end_mem_rss',
            'start_mem_vms',
            'end_mem_vms',
            'start_mem_uss',
            'end_mem_uss',
        ])
        # save the dataframe to csv file
        self.benchmark_df.to_csv(self.filename)

    def on_epoch_begin(self, batch, logs={}):
        self.epoch_time = -time.time()
        self.start_cpu_percent = self.process.cpu_percent()
        mem_info = self.process.memory_full_info()
        self.start_mem_rss = mem_info[0] / float(2 ** 20)
        self.start_mem_vms = mem_info[1] / float(2 ** 20)
        self.start_mem_uss = mem_info[3] / float(2 ** 20)

    def on_epoch_end(self, batch, logs={}):
        self.epoch_time += time.time()
        self.end_cpu_percent = self.process.cpu_percent()
        mem_info = self.process.memory_full_info()
        self.end_mem_rss = mem_info[0] / float(2 ** 20)
        self.end_mem_vms = mem_info[1] / float(2 ** 20)
        self.end_mem_uss = mem_info[3] / float(2 ** 20)

        self.benchmark_df.loc[len(self.benchmark_df.index)] = [
            self.epoch_time,
            self.start_cpu_percent,
            self.end_cpu_percent,
            self.start_mem_rss,
            self.end_mem_rss,
            self.start_mem_vms,
            self.end_mem_vms,
            self.start_mem_uss,
            self.end_mem_uss,
        ]

        self.benchmark_df.to_csv(self.filename)



"""
Show how to use DDP, Horovod and DeepSpeed strategies interchangeably
with an extremely simple neural network.
"""
from typing import Dict
import argparse
import time

import torch
from torch import nn
from torch.utils.data import Dataset

import horovod.torch as hvd

from itwinai.torch.distributed import (
    distributed_resources_available,
    TorchDistributedStrategy,
    TorchDDPStrategy,
    HorovodStrategy,
    DeepSpeedStrategy,
    NonDistributedStrategy
)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--strategy", "-s", type=str,
        choices=['ddp', 'horovod', 'deepspeed'],
        default='ddp'
    )
    parser.add_argument(
        "--shuffle_dataloader",
        action=argparse.BooleanOptionalAction
    )
    parser.add_argument(
        '--batch-size', type=int, default=10,
        help='input batch size for training (default: 10)')

    # DeepSpeed: needs to be removed
    import deepspeed
    parser.add_argument('--local_rank', type=int, default=-1,
                        help='local rank passed from distributed launcher')
    parser = deepspeed.add_config_arguments(parser)
    args = parser.parse_args()
    return args


class UniformRndDataset(Dataset):
    """Dummy torch dataset."""

    def __init__(self, x_size: int, y_size: int, len: int = 100):
        super().__init__()
        self.x_size = x_size
        self.y_size = y_size
        self.len = len

    def __len__(self):
        return self.len

    def __getitem__(self, index):
        return torch.rand(self.x_size), torch.rand(self.y_size)


def training_fn(
        args: argparse.Namespace,
        strategy: TorchDistributedStrategy,
        distribute_kwargs: Dict
) -> int:
    """Dummy training function."""
    strategy.init()

    # Local model
    model = nn.Linear(3, 4)
    optim = torch.optim.Adam(model.parameters(), lr=1e-3)
    loss_fn = nn.MSELoss()
    # Distributed model
    model, optim, lr_sched = strategy.distributed(
        model, optim, lr_scheduler=None, **distribute_kwargs
    )

    # Dataset
    train_set = UniformRndDataset(x_size=3, y_size=4)
    # Distributed dataloader
    train_loader = strategy.create_dataloader(
        train_set,
        batch_size=args.batch_size,
        num_workers=1,
        shuffle=True
    )

    # Device allocated for this worker
    device = strategy.device()

    for epoch in range(2):
        # IMPORTANT: set current epoch ID in distributed sampler
        train_loader.sampler.set_epoch(epoch)

        for (x, y) in train_loader:
            # print(f"tensor to cuda:{device}")
            x = x.to(device)
            y = y.to(device)

            optim.zero_grad()

            y_pred = model(x)

            loss = loss_fn(y_pred, y)
            loss.backward()

            optim.step()

            if strategy.is_main_worker:
                print(f"Loss [epoch={epoch}]: {loss.item()}")
            # print(f"NNLoss [epoch={epoch}]: {loss.item()}")

        # Update scheduler
        if lr_sched:
            lr_sched.step()

    time.sleep(1)
    print(f"<Global rank: {strategy.global_rank()}> - TRAINING FINISHED")
    strategy.clean_up()
    return 123


if __name__ == "__main__":

    args = parse_args()

    # Instantiate Strategy
    if not distributed_resources_available():
        print("WARNING: falling back to non-distributed strategy.")
        strategy = NonDistributedStrategy()
        distribute_kwargs = {}
    elif args.strategy == 'ddp':
        strategy = TorchDDPStrategy(backend='nccl')
        distribute_kwargs = {}
    elif args.strategy == 'horovod':
        strategy = HorovodStrategy()
        distribute_kwargs = dict(
            compression=hvd.Compression.none,
            op=hvd.Average,
            gradient_predivide_factor=1.0
        )
    elif args.strategy == 'deepspeed':
        strategy = DeepSpeedStrategy(backend='nccl')
        distribute_kwargs = dict(
            config_params=dict(train_micro_batch_size_per_gpu=args.batch_size)
        )
    else:
        raise NotImplementedError(
            f"Strategy {args.strategy} is not recognized/implemented.")
    # Launch distributed training
    training_fn(args, strategy, distribute_kwargs)
